"""Pipeline to retrieve GenBank records using accessions stored
in the MySQL database.
"""

import argparse
import re
import time
from pathlib import Path

from Bio import SeqIO

from pdm_utils.functions import configfile
from pdm_utils.functions import fileio
from pdm_utils.functions import ncbi
from pdm_utils.functions import mysqldb
from pdm_utils.functions import pipelines_basic

# GLOBAL VARIABLES
# -----------------------------------------------------------------------------
DEFAULT_FOLDER_NAME = f"{time.strftime('%Y%m%d')}_gb_records"
FILTER_KEY = "phage"

RECORD_FILE_TYPES = ["gb", "tbl"]


# MAIN FUNCTIONS
# -----------------------------------------------------------------------------
# TODO unittest.
def main(unparsed_args_list):
    """Run main get_gb_records pipeline."""
    # Parse command line arguments
    args = parse_args(unparsed_args_list)

    # Create config object with data obtained from file and/or defaults.
    config = configfile.build_complete_config(args.config_file)

    alchemist = pipelines_basic.build_alchemist(args.database, config=config)
    mysqldb.check_schema_compatibility(alchemist.engine,
                                       "the get_gb_records pipeline")

    values = pipelines_basic.parse_value_input(args.input)

    execute_get_gb_records(alchemist, args.file_type,
                           folder_path=args.folder_path, 
                           folder_name=args.folder_name, 
                           config=config,
                           values=values, verbose=args.verbose, 
                           filters=args.filters, groups=args.groups) 


# TODO unittest.
def parse_args(unparsed_args_list):
    """Parses export_db arguments and stores them with an argparse object.

    :param unparsed_args_list: Input a list of command line args.
    :type unparsed_args_list: list[str]
    :returns: ArgParse module parsed args.
    """
    DATABASE_HELP = "Name of the MySQL database to export from."

    FORCE_HELP = """
        Get gb record option that aggresively creates and overwrites 
        directories.
        """
    CONFIG_FILE_HELP = """
        Get gb record option that enables use of a config file for sourcing
        credentials
            Follow selection argument with the path to the config file
            specifying MySQL and NCBI credentials.
        """
    VERBOSE_HELP = """
        Get gb record option that enables progress print statements.
        """
    FOLDER_PATH_HELP = """
        Get gb record option to change the path
        of the directory where the exported files are stored.
            Follow selection argument with the path to the
            desired export directory.
        """
    FOLDER_NAME_HELP = """
        Export option to change the name
        of the directory where the exported files are stored.
            Follow selection argument with the desired name.
        """

    FILE_TYPE_HELP = """
        Get gb record option to select the file type retrieved from GenBank
            Follow selection argument with the desired file type.
        """

    IMPORT_FILE_HELP = """
        Selection input option that imports values from a csv file.
            Follow selection argument with path to the
            csv file containing the names of each genome in the first column.
        """
    SINGLE_GENOMES_HELP = """
        Selection input option that imports values from cmd line input.
            Follow selection argument with space separated
            names of genomes in the database.
        """
    TABLE_HELP = """
        Selection option that changes the table in the database selected from.
            Follow selection argument with a valid table from the database.
        """
    WHERE_HELP = """
        Data filtering option that filters data by the inputted expressions.
            Follow selection argument with formatted filter expression:
                {Table}.{Column}={Value}
        """
    GROUP_BY_HELP = """
        Data selection option that groups data by the inputted columns.
            Follow selection argument with formatted column expressions:
                {Table}.{Column}={Value}
        """
    parser = argparse.ArgumentParser()
    parser.add_argument("database", type=str, help=DATABASE_HELP)

    parser.add_argument("-c", "--config_file", 
                                type=pipelines_basic.convert_file_path,
                                help=CONFIG_FILE_HELP)
    parser.add_argument("-m", "--folder_name",
                                type=str, help=FOLDER_NAME_HELP)
    parser.add_argument("-o", "--folder_path", 
                                type=Path,
                                help=FOLDER_PATH_HELP)
    parser.add_argument("-v", "--verbose", action="store_true",
                                help=VERBOSE_HELP)

    parser.add_argument("-ft", "--file_type", choices=RECORD_FILE_TYPES,
                                type=str, 
                                help=FILE_TYPE_HELP)
 
    parser.add_argument("-if", "--import_file", dest="input",
                                type=pipelines_basic.convert_file_path,
                                help=IMPORT_FILE_HELP)
    parser.add_argument("-in", "--import_names", nargs="*", dest="input",
                                help=SINGLE_GENOMES_HELP)
    parser.add_argument("-f", "--where", nargs="?", dest="filters",
                                help=WHERE_HELP)
    parser.add_argument("-g", "--group_by", nargs="*", dest="groups",
                                help=GROUP_BY_HELP)

    parser.set_defaults(file_type="gb",
                        folder_name=DEFAULT_FOLDER_NAME,
                        folder_path=None,
                        verbose=False, input=[],
                        filters="", groups=[], sort=[])

    parsed_args = parser.parse_args(unparsed_args_list[2:])    
    return parsed_args


def execute_get_gb_records(alchemist, file_type, folder_path=None,
                           folder_name=DEFAULT_FOLDER_NAME, 
                           config=None, values=None, verbose=False, 
                           force=False, filters="", groups=[]):
    """Executes the entirety of the get_gb_records pipeline

    :param alchemist: A connected and fully build AlchemyHandler object.
    :type alchemist: AlchemyHandler
    :param folder_path: Path to a valid dir for new dir creation.
    :type folder_path: Path
    :param folder_name: A name for the export folder.
    :type folder_name: str
    :param file_type: File type to be exported.
    :type file_type: str
    :param config: ConfigParser object containing NCBI credentials.
    :type config: ConfigParser
    :param force: A boolean to toggle aggresive building of directories.
    :type force: bool
    :param values: List of values to filter database results.
    :type values: list[str]
    :param verbose: A boolean value to toggle progress print statemtns.
    :type verbose: bool
    :param filters: A List of lists with filter value,grouped by ORs.
    :type filter: str
    :param groups: A list of supported MySQL column names to goup by.
    :type groups: list[str]
    """
    ncbi_creds = {}
    if config is not None:
        ncbi_creds = config["ncbi"]

    db_filter = pipelines_basic.build_filter(alchemist, FILTER_KEY, filters,
                                             values=values, verbose=verbose)

    if verbose:
        print("Creating records folder...")
    records_path = pipelines_basic.create_working_path(folder_path,
                                                       folder_name,
                                                       force=force)

    conditionals_map = pipelines_basic.build_groups_map(
                                                db_filter, records_path,
                                                groups=groups, verbose=verbose,
                                                force=force)

    values = db_filter.values
    for mapped_path in conditionals_map.keys():
        db_filter.reset()
        db_filter.values = values

        conditionals = conditionals_map[mapped_path]
        db_filter.values = db_filter.build_values(where=conditionals)

        # Create data sets
        if verbose:
            print("Retrieving accessions from the database...")
        accession_data = db_filter.select(["phage.PhageID", "phage.Accession"])

        acc_id_dict = {}
        for data_dict in accession_data:
            accession = data_dict["Accession"]
            if not (accession is None or accession == ""):
                acc_id_dict[accession] = data_dict["PhageID"]

        pipelines_basic.create_working_dir(mapped_path, force=force)
        if len(acc_id_dict.keys()) > 0:
            ncbi_handle = ncbi.get_verified_data_handle(
                                                     acc_id_dict,
                                                     ncbi_cred_dict=ncbi_creds,
                                                     file_type=file_type)

            copy_gb_data(ncbi_handle, acc_id_dict, mapped_path, file_type,
                         verbose=verbose)
        else:
            print(f"There are no records to retrieve for '{mapped_path}'.")
            continue


# TODO test.
def copy_gb_data(ncbi_handle, acc_id_dict, records_path, file_type,
                 verbose=False):
    """Save retrieved records to file."""
    if file_type == "gb":
        record_parser = SeqIO.parse(ncbi_handle, "genbank")

        seqrecords = []
        for record in record_parser:
            accession = record.annotations["accessions"][0]
            phage_id = acc_id_dict[accession]
            record.name = phage_id
            seqrecords.append(record)

        ncbi_handle.close()
        fileio.write_seqrecords(seqrecords, "gb", records_path,
                                verbose=verbose)

    elif file_type == "tbl":
        file_lines = ncbi_handle.readlines()

        feature_format = re.compile(">Feature ..\|(\w+)(\..)\|\n")

        file_handle = None
        for line in file_lines:
            if not re.match(feature_format, line) is None:
                accession_split = re.split(feature_format, line)
                accession = accession_split[1]
                phage_id = acc_id_dict[accession_split[1]]

                file_name = (f"{phage_id}.tbl")
                file_path = records_path.joinpath(file_name)
                file_handle = file_path.open(mode="w")

                file_handle.write(line)
            else:
                if file_handle is None:
                    continue
                file_handle.write(line)

        ncbi_handle.close()



"""Misc. functions to utilizes server."""

import getpass
import paramiko
from pdm_utils.functions import basic


# TODO unittest (but manually tested).
def set_log_file(filepath):
    """Set the filepath used to stored the Paramiko output.

    This is a soft requirement for compliance with Paramiko standards.
    If it is not set, paramiko throws an error.

    :param filepath: Path to file to log Paramiko results.
    :type filepath: Path
    """
    paramiko.util.log_to_file(filepath)


# TODO unittest (but manually tested).
def get_transport(host):
    """Create paramiko Transport with the server name.

    :param host: Server to connect to.
    :type host: str
    :returns:
        Paramiko Transport object. If the server is not available,
        None is returned.
    :rtype: Transport
    """
    # If the host name is not valid, paramiko is unable to find the host
    # and throws an error.
    try:
        transport = paramiko.Transport(host)
    except:
        transport = None
        print("Unable to find server: " + host)
    return transport


# TODO unittest (but manually tested).
def setup_sftp_conn(transport, user=None, pwd=None, attempts=1):
    """Get credentials and setup connection to the server.

    :param transport: Paramiko Transport object directed towards a valid server.
    :type transport: Transport
    :param attempts: Number of attempts to connect to the server.
    :type attempts: int
    :returns:
        Paramiko SFTPClient connection. If no connection can be made,
        None is returned.
    :rtype: SFTPClient
    """
    sftp = None
    # Note: the way paramiko manages the connectionn, this loop
    # doesn't seem to work with multiple attempts.
    # There may be a transport attribute that needs to be reset or
    # a new transport object may need to be generated after each failed
    # attempt.

    # Only give one attempt if user and password is provided as parameter.
    if user is not None and pwd is not None:
        attempts = 1

    while (attempts > 0 and sftp is None):
        if user is None or pwd is None:
            user, pwd = basic.get_user_pwd(user_prompt="Server username: ",
                                           pwd_prompt="Server password: ")
        try:
            transport.connect(username=user, password=pwd)
        except:
            print("Unable to connect to server. "
                  "Incorrect username and password")
            # If user and pwd is incorrect, reset to None so that
            # user is prompted to provide them again (if attempts > 0)
            user = None
            pwd = None
            attempts -= 1
        else:
            sftp = paramiko.SFTPClient.from_transport(transport)
    return sftp

# TODO unittest (but manually tested).
def upload_file(sftp, local_filepath, remote_filepath):
    """Upload a file to the server.

    :param sftp: Paramiko SFTPClient connection to a server.
    :type sftp: SFTPClient
    :param local_filepath: Absoluate path to file to be uploaded.
    :type local_filepath: str
    :param remote_filepath: Absoluate path to server destination.
    :type remote_filepath: str
    :returns: Indicates whether upload was successful.
    :rtype: bool
    """
    try:
        sftp.put(local_filepath, remote_filepath)
        result = True
    except:
        print("Problems encountered with the "
              "specified local or destination directories.")
        result = False
    return result



import csv
import shutil
import sys
from functools import singledispatch
from pathlib import Path

from pdm_utils.classes.alchemyhandler import AlchemyHandler
from pdm_utils.classes.filter import Filter
from pdm_utils.functions import basic


# PIPELINE OUTPUT HANDLING
# -----------------------------------------------------------------------------
def create_working_dir(working_path, dump=False, force=False):
    if not dump:
        if not force:
            if working_path.is_dir():
                print("COWARDLY ABORTING PIPELINE: "
                      f"Directory '{working_path}' already exists.")
                sys.exit(1)

    working_path.mkdir(parents=True, exist_ok=True)


def create_working_path(folder_path, folder_name, dump=False, force=False,
                        attempt=50):
    if folder_path is None:
        working_path = create_default_path(folder_name, force=force,
                                           attempt=attempt)
    else:
        working_path = folder_path.joinpath(folder_name)

    if dump:
        working_path = working_path.parent

    return working_path


def create_default_path(name, force=False, attempt=50):
    default_path = Path.cwd().joinpath(name)
    if not force:
        default_path = basic.make_new_dir(Path.cwd(), default_path,
                                          attempt=attempt, mkdir=False)

    return default_path


def convert_dir_path(path):
    """Function to convert argparse input to a working directory path.

    :param path: A string to be converted into a Path object.
    :type path: str
    :returns: A Path object converted from the inputed string.
    :rtype: Path
    """
    return basic.set_path(Path(path), kind="dir")


def convert_file_path(path):
    """Function to convert argparse input to a working file path.

    :param path: A string to be converted into a Path object.
    :type path: str
    :returns: A Path object converted from the inputed string.
    :rtype: Path
    """
    return basic.set_path(Path(path), kind="file")


# PIPELINE INPUT HANDLING
# -----------------------------------------------------------------------------
@singledispatch
def parse_value_input(value_list_input):
    """Function to convert values input to a recognized data types.

    :param value_list_input: Values stored in recognized data types.
    :type value_list_input: list[str]
    :type value_list_input: Path
    :returns: List of values to filter database results.
    :rtype: list[str]
    """

    print("Value list input is of an unexpected type.")
    sys.exit(1)


@parse_value_input.register(Path)
def _(value_list_input):
    value_list = []
    with open(value_list_input, newline='') as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        for name in csv_reader:
            value_list.append(name[0])
    return value_list


@parse_value_input.register(list)
def _(value_list_input):
    return value_list_input


# MYSQL FILTERS AND HANDLER FUNCTIONS
def build_alchemist(database, ask_database=True, config=None, dialect="mysql"):
    if config is not None:
        username = config["mysql"].get("user")
        password = config["mysql"].get("password")
        if not (username is None or password is None):
            alchemist = AlchemyHandler(username=username, password=password,
                                       dialect=dialect)
            alchemist.connect(login_attempts=0, pipeline=True)

            alchemist.database = database
            alchemist.connect(ask_database=ask_database, pipeline=True)

            return alchemist

    alchemist = AlchemyHandler(database=database)
    alchemist.connect(ask_database=ask_database, pipeline=True)

    return alchemist


def build_filter(alchemist, key, filters, values=None, verbose=False):
    """Applies MySQL WHERE clause filters using a Filter.

    :param alchemist: A connected and fully built AlchemyHandler object.
    :type alchemist: AlchemyHandler
    :param table: MySQL table name.
    :type table: str
    :param filters: A list of lists with filter values, grouped by ORs.
    :type filters: list[list[str]]
    :param groups: A list of supported MySQL column names.
    :type groups: list[str]
    :returns: filter-Loaded Filter object.
    :rtype: Filter
    """
    db_filter = Filter(alchemist=alchemist)
    db_filter.key = key
    db_filter.values = values

    if filters != "":
        try:
            db_filter.add(filters)
        except:
            print("Please check your syntax for the conditional string: "
                  f"{filters}")
            exit(1)

        db_filter.parenthesize()

    return db_filter


def add_sort_columns(db_filter, sort_columns, verbose=False):
    if verbose:
        print("Processing columns for sorting...")

    try:
        db_filter.sort(sort_columns)
    except:
        print("Please check your syntax for sorting columns:\n"
              f"{', '.join(sort_columns)}")
        sys.exit(1)


def build_groups_map(db_filter, export_path, groups=[], verbose=False,
                     force=False, dump=False):
    """Function that generates a map between conditionals and grouping paths.

    :param db_filter: A connected and fully loaded Filter object.
    :type db_filter: Filter
    :param export_path: Path to a dir for new dir creation.
    :type folder_path: Path
    :param groups: A list of supported MySQL column names.
    :type groups: list[str]
    :param conditionals_map: A mapping between group conditionals and Paths.
    :type conditionals_map: dict{Path:list}
    :param verbose: A boolean value to toggle progress print statements.
    :type verbose: bool
    :param previous: Value set by function to provide info for print statements
    :type previous: str
    :param depth: Value set by function to provide info for print statements.
    :type depth: int
    :returns conditionals_map: A mapping between group conditionals and Paths.
    :rtype: dict{Path:list}
    """
    conditionals_map = {}
    try:
        build_groups_tree(db_filter, export_path, conditionals_map,
                          groups=groups, verbose=False, force=False)
    except:        
        print("COWARDLY ABORTING PIPELINE: "
              "Found duplicate directories during path structuring.")
        sys.exit(1)

    return conditionals_map


def build_groups_tree(db_filter, export_path, conditionals_map, groups=[],
                      verbose=False, force=False, previous=None, depth=0):
    """Recursive function that generates directories based on groupings.

    :param db_filter: A connected and fully loaded Filter object.
    :type db_filter: Filter
    :param export_path: Path to a dir for new dir creation.
    :type folder_path: Path
    :param groups: A list of supported MySQL column names.
    :type groups: list[str]
    :param conditionals_map: A mapping between group conditionals and Paths.
    :type conditionals_map: dict{Path:list}\
    :param verbose: A boolean value to toggle progress print statements.
    :type verbose: bool
    :param previous: Value set by function to provide info for print statements
    :type previous: str
    :param depth: Value set by function to provide info for print statements.
    :type depth: int
    :returns conditionals_map: A mapping between group conditionals and Paths.
    :rtype: dict{Path:list}
    """
    groups = groups.copy()
    conditionals = db_filter.build_where_clauses()
    if not groups:
        conditionals_map.update({export_path: conditionals})
        return

    current_group = groups.pop(0)
    if verbose:
        if depth > 0:
            dots = ".." * depth
            print(f"{dots}Grouping by {current_group} in {previous}...")
        else:
            print(f"Grouping by {current_group}...")

    try:
        group_column = db_filter.get_column(current_group)
    except:
        print(f"Group '{current_group}' is not a valid group.")
        sys.exit(1)

    transposed_values = db_filter.build_values(column=group_column,
                                               where=conditionals)

    for group in transposed_values:
        group_path = export_path.joinpath(str(group))
        if group_path.is_dir():
            if force:
                shutil.rmtree(group_path)
            else:
                raise OSError("COWARDLY ABORTING PIPELINE: "
                              "Found pre-existing directories during "
                              "group path stucturing.")

        db_filter_copy = db_filter.copy()
        db_filter_copy.add(f"{current_group}={group}")

        previous = f"{current_group} {group}"
        build_groups_tree(db_filter_copy, group_path, conditionals_map,
                          groups=groups, verbose=verbose,
                          force=force, previous=previous, depth=depth+1)



import re
from collections import OrderedDict

from Bio.Seq import Seq
from Bio.SeqFeature import SeqFeature, CompoundLocation, FeatureLocation
from Bio.SeqRecord import SeqRecord

#GLOBAL VARIABLES
#-----------------------------------------------------------------------------
TBL_HEADER_FORMAT = re.compile(">Feature (\w+)\|(\w+)\.(.)\|\n") 
TBL_FEATURE_FORMAT = re.compile("(\d+)\t(\d+)\t(\w+)\n")
TBL_REGIONS_FORMAT = re.compile("(\d+)\t(\d+)\n")
TBL_QUALIFIER_FORMAT = re.compile("\t\t\t(.+)\t(.+)\n")
TBL_OPEN_QUALIFIER_FORMAT = re.compile("\t\t\t(.+)\n")

class FeatureTableParser:
    """Class to act as a generator for reading (five-column) feature tables
    and retrieving Biopython SeqRecord objects.
    """
    def __init__(self, filehandle):
        self.filehandle = filehandle
        self._prev_line = ""

    #For interfacing with python's builtin library
    def __next__(self):
        return self.next()
    
    def next(self):
        if self.filehandle.closed:
            raise IOError("Access to file has been closed unexpectedly.")

        #Checks if file previously parsed into the start of a possible table
        if parse_tbl_data_type(self._prev_line) == "header":
            tbl_data = [self._prev_line]
        else:
            tbl_data = []
            #Reads through the whole file for a feature table header
            for line in self.filehandle:
                self._prev_line = line
                if parse_tbl_data_type(line) != "header":
                    continue
                else:
                    tbl_data = [line]
                    break
           
            #If no header is detected, then tbl_data is empty.
            if not tbl_data:
                raise StopIteration("File contains no more Feature Tables")

        for line in self.filehandle:
            self._prev_line = line
            data_type = parse_tbl_data_type(line)
            if not(data_type == "header" or data_type is None):
                tbl_data.append(line)
            else:
                break

        record = convert_tbl_data_to_record(tbl_data)
        return record

    def __iter__(self):
        return self

def convert_tbl_data_to_record(tbl_data):
    """Converts string lines from a five_column feature table to a seqrecord.
    
    :param tbl_data: A list of the lines of data from a feature table file.
    :type tbl_data: list
    :returns: Returns a Biopython SeqRecord object loaded with given data.
    :rtype: SeqRecord
    """
    #Tbl_data is expected to contain the header as the first value.
    header = tbl_data[0]
    if parse_tbl_data_type(header) != "header":
        raise ValueError("Five column table data does not have a proper header")

    #Capturing groups in the regular expression capture:
        #[1] The feature table accession prefix
        #[2] The accession for the feature table genome
        #[3] The version attributed to the feature table genome
    header_split = re.split(TBL_HEADER_FORMAT, header) 
    record = SeqRecord(Seq(''))
    record.id = header_split[2] 
    record.annotations = {"sequence_version" : str(header_split[3]),
                          "tbl_prefix"       : str(header_split[1])}

    if len(tbl_data) == 1:
        return record

    features = []
    qualifiers = OrderedDict()
    coordinates = []
    last_data_type = None
    feature_split = []
    for i in range(len(tbl_data[1:])):
        line = tbl_data[i+1]
        data_type = parse_tbl_data_type(line)

        #Tbl_data is not expected to have another header
        if data_type == "header":
            raise SyntaxError("".join([
                       f"Found table header within data on line {i+2} "
                        "of data.\n",
                        line.rstrip("\n")]))
        #Tbl_data feature can have multiple regions, so coordinates are saved
        #and upon reaching the next feature are appended to a feature list
        #with the respective qualifiers (and regions)
        elif data_type == "feature":
            if coordinates:
                feature = feature_data_to_seqfeature(coordinates,
                                                        feature_split[3])
                feature.qualifiers = qualifiers
                features.append(feature)
            coordinates = []
            qualifiers = OrderedDict()

            feature_split = re.split(TBL_FEATURE_FORMAT, line)
            coordinates.append((int(feature_split[1]), int(feature_split[2])))
        elif data_type == "region":
            if not last_data_type in ["feature", "region"]:
                raise SyntaxError("".join([
                        "Five column feature table not formatted correctly.\n"
                       f"Orphaned multiple region data found on line {i+2}.\n",
                        line.rstrip("\n")]))

            regions_split = re.split(TBL_REGIONS_FORMAT, line)
            coordinates.append((int(regions_split[1]), int(regions_split[2])))
        elif data_type == "qualifier":
            qualifier_split = re.split(TBL_QUALIFIER_FORMAT, line)

            qualifier_storage = qualifiers.get(qualifier_split[1])
            if qualifier_storage:
                qualifiers[qualifier_split[1]].append(qualifier_split[2])
            else:
                qualifiers[qualifier_split[1]] = [qualifier_split[2]]
        elif data_type == "open_qualifier":
            open_qualifier_split = re.split(TBL_OPEN_QUALIFIER_FORMAT, line)
            qualifiers[open_qualifier_split[1]] = [""]
        else:
            raise SyntaxError("".join([
                        "Five column feature table not formatted correctly.\n"
                       f"Unrecognized data format on line {i+1} of data.\n",
                        line.rstrip("\n")]))

        last_data_type = data_type

    #To capture the last feature
    if coordinates:
        feature = feature_data_to_seqfeature(coordinates,
                                                feature_split[3])
        feature.qualifiers = qualifiers
        features.append(feature)

    record.features = features
    return record

def feature_data_to_seqfeature(coordinates, feature_type):
    """Converts coordinates received from feature table parsing to a SeqFeature.

    :param coordinates: Start and end positions for the feature.
    :type coordinates: list[tuple(int, int)]
    :param feature_type: Label of the parsed feature.
    :type str
    :returns: Returns a Biopython SeqFeature loaded with given coordinates.
    :rtype: SeqFeature
    """
    locations = []
    for coordinate in coordinates:
        if coordinate[0] <= coordinate[1]:
            start = int(coordinate[0]) - 1
            stop = int(coordinate[1])
            strand = 1
        else:
            start = int(coordinate[1]) - 1
            stop = int(coordinate[0])
            strand = -1
        
        locations.append(FeatureLocation(start, stop, strand=strand))

    if len(locations) == 1:
        feature = SeqFeature(locations[0], type=feature_type)
    else:
        feature = SeqFeature(CompoundLocation(locations), type=feature_type)

    return feature

def parse_tbl_data_type(data_line):
    """Parses a five-column table data line and returns it's structure type.

    :param data_line: Line of data from a five-column feature table.
    :type data_line: str
    :returns: Returns the type of data line the line is from the table.
    :rtype: str
    """
    data_type = None
    if not re.match(TBL_HEADER_FORMAT, data_line) is None:
        data_type = "header"
    elif not re.match(TBL_FEATURE_FORMAT, data_line) is None:
        data_type = "feature"
    elif not re.match(TBL_REGIONS_FORMAT, data_line) is None:
        data_type = "region"
    elif not re.match(TBL_QUALIFIER_FORMAT, data_line) is None:
        data_type = "qualifier"
    elif not re.match(TBL_OPEN_QUALIFIER_FORMAT, data_line) is None:
        data_type = "open_qualifier"
   
    return data_type 




"""Use this script to run all pipelines within the pipelines folder.
It verifies a valid pipeline is selected,
then passes all command line arguments to the main pipeline module.
"""
import sys
import argparse

from pdm_utils.pipelines import compare_db
from pdm_utils.pipelines import convert_db
from pdm_utils.pipelines import export_db
from pdm_utils.pipelines import find_domains
from pdm_utils.pipelines import freeze_db
from pdm_utils.pipelines import get_data
from pdm_utils.pipelines import get_db
from pdm_utils.pipelines import get_gb_records
from pdm_utils.pipelines import import_genome
from pdm_utils.pipelines import phamerate
from pdm_utils.pipelines import pham_finder
from pdm_utils.pipelines import push_db
from pdm_utils.pipelines import revise
from pdm_utils.pipelines import pham_review
from pdm_utils.pipelines import update_field

VALID_PIPELINES = {"compare", "convert", "export", "find_domains",
                   "find_phams", "freeze", "get_data", "get_db",
                   "get_gb_records", "import", "phamerate", "push",
                   "revise", "pham_review", "update"}


def parse_args(unparsed_args):
    """
    Use argparse to verify pipeline argument only.

    :param unparsed_args: raw command line args
    :type unparsed_args: list
    """
    run_help = "Commandline script to call a pdm_utils pipeline."
    usage = "python3 -m pdm_utils [pipeline]"
    pipeline_help = "name of the pdm_utils pipeline to run"

    parser = argparse.ArgumentParser(description=run_help, usage=usage)
    parser.add_argument("pipeline", type=str, choices=list(VALID_PIPELINES),
                        help=pipeline_help)

    # Assumed command line arg structure:
    # python3 -m pdm_utils.run <pipeline> <additional args...>
    # sys.argv:      [0]            [1]         [2...]
    return parser.parse_args(unparsed_args[1:2])


def main(unparsed_args=None):
    """Commandline entry point for the pdm_utils package."""
    if not unparsed_args:
        if len(sys.argv) == 1:
            sys.argv.append("-h")
        unparsed_args = sys.argv

    args = parse_args(unparsed_args)

    if args.pipeline == "compare":
        compare_db.main(unparsed_args)
    elif args.pipeline == "convert":
        convert_db.main(unparsed_args)
    elif args.pipeline == "export":
        export_db.main(unparsed_args)
    elif args.pipeline == "find_domains":
        find_domains.main(unparsed_args)
    elif args.pipeline == "find_phams":
        pham_finder.main(unparsed_args)
    elif args.pipeline == "freeze":
        freeze_db.main(unparsed_args)
    elif args.pipeline == "get_data":
        get_data.main(unparsed_args)
    elif args.pipeline == "get_db":
        get_db.main(unparsed_args)
    elif args.pipeline == "get_gb_records":
        get_gb_records.main(unparsed_args)
    elif args.pipeline == "import":
        import_genome.main(unparsed_args)
    elif args.pipeline == "phamerate":
        phamerate.main(unparsed_args)
    elif args.pipeline == "push":
        push_db.main(unparsed_args)
    elif args.pipeline == "revise":
        revise.main(unparsed_args)
    elif args.pipeline == "pham_review":
        pham_review.main(unparsed_args)
    elif args.pipeline == "update":
        update_field.main(unparsed_args)
    else:
        pass
    print("\n\nPipeline completed")



"""Pipeline to compare data between MySQL, PhagesDB, and GenBank databases."""

# TODO this pipeline is not fully integrated into the pdm_utils package, and
# several parts need to be improved.

# 1. It originally utilized its own Cds and Genome classes. It now relies on
# the base pdm_utils Cds and Genome classes, but then requires additional
# attributes and methods, which are added in this module. The pipeline keeps
# track of errors using booleans and tallies, in contrast to the import pipeline
# which keeps track of errors using Evaluation objects. Ultimately, this pipeline
# should probably realy on Evaluation objects.

# 2. It also utilizes GenomeTriad, CdsPair, and DbCompareSummary classes
# which need to be substantially refactored and generalized.

# 3. In order to compare PhagesDB to MySQL, it needs to retrieve genome
# sequences from PhagesDB. This requires retrieving and parsing thousands of
# fasta files, which is very slow. It does this for all genomes on PhagesDB,
# even if they are not matched to MySQL genomes, so it is a big waste of time.

# 4. The data consistency checks are not comprehensive. Many more checks
# should be added.

# 5. None of the functions in this pipeline are tested.

# Note this script compares and matches data from PhagesDB, GenBank, and MySQL.
# As a result, there are many similarly named variables.
# Variables are prefixed to indicate database:
# GenBank =  "gbk", "g"
# MySQL = "mysql", "m"
# PhagesDB = "pdb", "p"

import argparse
import csv
from datetime import date
import os
import pathlib
import re
import sys
import time

from Bio import SeqIO, Entrez
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from Bio.Alphabet import IUPAC

from pdm_utils.classes import cds
from pdm_utils.classes import genome
from pdm_utils.classes import cdspair
from pdm_utils.classes import dbcomparesummary
from pdm_utils.classes import genometriad
from pdm_utils.classes.alchemyhandler import AlchemyHandler
from pdm_utils.classes.filter import Filter
from pdm_utils.constants import constants
from pdm_utils.functions import basic
from pdm_utils.functions import configfile
from pdm_utils.functions import flat_files
from pdm_utils.functions import mysqldb
from pdm_utils.functions import mysqldb_basic
from pdm_utils.functions import ncbi
from pdm_utils.functions import parsing
from pdm_utils.functions import phagesdb
from pdm_utils.functions import querying

DEFAULT_OUTPUT_FOLDER = os.getcwd()

CURRENT_DATE = date.today().strftime("%Y%m%d")
WORKING_FOLDER = f"{CURRENT_DATE}_compare"

GENBANK_OUTPUT = "gbk_records"
MYSQL_OUTPUT = "mysql_records"
PHAGESDB_OUTPUT = "phagesdb_records"

ERROR_FOLDER = "errors"
RECORD_FOLDER = "records"
RESULTS_FOLDER = "results"

COMPARE_SETTINGS = "compare_settings.csv"
DUPLICATE_MYSQL_NAMES = "mysql_name_duplicates.csv"
DUPLICATE_MYSQL_ACC = "mysql_accession_duplicates.csv"
DUPLICATE_PDB_NAMES = "phagesdb_name_duplicates.csv"
FAILED_ACC_RETRIEVE = "mysql_genomes_unmatched_to_genbank.csv"
UNMATCHED_GENOMES = "mysql_genomes_unmatched_to_pdb.csv"

GNM_SUMMARY_OUTPUT = "summary_genome.csv"
CDS_SUMMARY_OUTPUT = "summary_cds.csv"

GENOME_OUTPUT = "genome.csv"
GENE_OUTPUT = "cds.csv"

PHAGE_QUERY = ("SELECT PhageID, Name, HostGenus, Sequence, Length, "
               "Status, Cluster, Accession, RetrieveRecord, "
               "DateLastModified, AnnotationAuthor FROM phage")
GENE_QUERY = ("SELECT PhageID, GeneID, Name, Start, Stop, Orientation, "
              "CONVERT(Translation USING utf8) as Translation, "
              "Notes from gene")
VERSION_QUERY = "SELECT Version FROM version"

TARGET_TABLE = "phage"

# Identifiers for object types based on their database source.
GNM_MYSQL = "mysql"
GNM_PDB = "phagesdb"
GNM_GBK = "genbank"
CDS_MYSQL = "cds_mysql"
CDS_GBK = "cds_genbank"
CDSPAIR_MYSQL_GBK = "mysql_genbank"



# TODO these Genome and Cds class modifications are a temporary
# solution to be able to integrate the compare pipeline into pdm_utils.
# These modifications should ultimately be
# integrated directly into the base pdm_utils classes or
# the compare pipeline should be revamped to not need these extra
# attributes and methods.

### Genome class modifications

# TODO refactor and test.
def modify_genome_class(GenomeClass):
    """Add new attributes and methods to Genome class."""

    # New attributes
    setattr(GenomeClass, "record_name", "")
    setattr(GenomeClass, "record_id", "")
    setattr(GenomeClass, "source_feature_organism", "")
    setattr(GenomeClass, "source_feature_host", "")
    setattr(GenomeClass, "source_feature_lab_host", "")
    setattr(GenomeClass, "_missing_locus_tags_tally", 0)
    setattr(GenomeClass, "_locus_tag_typos_tally", 0)
    setattr(GenomeClass, "_description_field_error_tally", 0)
    setattr(GenomeClass, "_status_accession_error", False)
    setattr(GenomeClass, "_status_description_error", False)
    setattr(GenomeClass, "_genes_with_errors_tally", 0)
    setattr(GenomeClass, "_nucleotide_errors", False)
    setattr(GenomeClass, "_cds_features_with_translation_error_tally", 0)
    setattr(GenomeClass, "_cds_features_boundary_error_tally", 0)

    # New methods
    setattr(GenomeClass, "compute_nucleotide_errors",
            compute_nucleotide_errors)
    setattr(GenomeClass, "compute_cds_feature_errors",
            compute_cds_feature_errors)
    setattr(GenomeClass, "check_status_accession",
            check_status_accession)
    setattr(GenomeClass, "compute_status_description_error",
            compute_status_description_error)
    setattr(GenomeClass, "compute_genes_with_errors_tally",
            compute_genes_with_errors_tally)
    setattr(GenomeClass, "compute_gbk_cds_feature_errors",
            compute_gbk_cds_feature_errors)


# Genome error checks

# TODO refactor and test.
def compute_nucleotide_errors(self, dna_alphabet):
    nucleotide_set = set(self.seq)
    nucleotide_error_set = nucleotide_set - dna_alphabet
    if len(nucleotide_error_set) > 0:
        self._nucleotide_errors = True

# TODO refactor and test.
def compute_cds_feature_errors(self):
    for cds_feature in self.cds_features:
        if cds_feature._amino_acid_errors:
            self._cds_features_with_translation_error_tally += 1
        if cds_feature._boundary_error:
            self._cds_features_boundary_error_tally += 1

# TODO refactor and test.
def check_status_accession(self):

    # Be sure to first set the accession attribute before the
    # annotation_status attribute, else this will throw an error.
    # Now that the AnnotationAuthor field contains authorship data, the
    # 'unknown' annotation annotation_status now reflects
    # an 'unknown' annotation (in regards to if it was auto-annotated
    # or manually annotated).
    # So for the annotation_status-accession error, if the
    # annotation_status is 'unknown', there is no reason to assume
    # whether there should be an accession or not. Only for 'final'
    # (manually annotated) genomes should there be an accession.
    if self.annotation_status == "final" and self.accession == "":
        self._status_accession_error = True

# TODO refactor and test.
def compute_status_description_error(self):
    # Iterate through all CDS features, see if they have descriptions,
    # then compare to the annotation_status.
    for feature in self.cds_features:
        if feature.raw_description != "":
            self._cds_descriptions_tally += 1
    if self.annotation_status == "draft" and self._cds_descriptions_tally > 0:
        self._status_description_error = True
    elif self.annotation_status == "final" and self._cds_descriptions_tally == 0:
        self._status_description_error = True
    else:
        pass


# TODO refactor and test.
def compute_genes_with_errors_tally(self):
    #Even though this method iterates through the CDS features
    # like the compute_status_description_error does,
    #it has to be kept separate, since you need to wait to run
    # this method after all genome and gene matching is completed.
    for feature in self.cds_features:
        # Need to first compute the number of errors per gene
        feature.check_for_errors()
        if feature._total_errors > 0:
            self._genes_with_errors_tally += 1

# TODO refactor and test.
def compute_gbk_cds_feature_errors(self):
    for cds_feature in self.cds_features:

        # counting descriptions should skip if it is blank
        # or "hypothetical protein".
        if cds_feature.product != "":
            self._cds_products_tally += 1

        if cds_feature.function != "":
            self._cds_functions_tally += 1

        if cds_feature.note != "":
            self._cds_notes_tally += 1

        if cds_feature._locus_tag_missing:
            self._missing_locus_tags_tally += 1
        else:

            search_name = basic.edit_suffix(self.name, "remove").lower()
            pattern4 = re.compile(search_name)

            search_result = pattern4.search(cds_feature.locus_tag.lower())

            if search_result == None:
                self._locus_tag_typos_tally += 1
                cds_feature.set_locus_tag_typo() # Sets to True

        if cds_feature._description_field_error:
            self._description_field_error_tally += 1


### Cds class modification

# TODO refactor and test.
def modify_cds_class(CdsClass):
    """Add new attributes and methods to Cds class."""

    setattr(CdsClass, "_search_genome_id", "")
    setattr(CdsClass, "_start_end_strand_id", "")
    setattr(CdsClass, "_end_strand_id", "")
    setattr(CdsClass, "_locus_tag_missing", False)
    setattr(CdsClass, "_locus_tag_typo", False)
    setattr(CdsClass, "_description_field_error", False)
    setattr(CdsClass, "_amino_acid_errors", False)
    setattr(CdsClass, "_boundary_error", False)
    setattr(CdsClass, "_unmatched_error", False)
    setattr(CdsClass, "_total_errors", 0)

    # New methods
    setattr(CdsClass, "set_start_end_strand_id", set_start_end_strand_id)
    setattr(CdsClass, "set_search_genome_id", set_search_genome_id)
    setattr(CdsClass, "check_locus_tag", check_locus_tag)
    setattr(CdsClass, "compute_amino_acid_errors", compute_amino_acid_errors)
    setattr(CdsClass, "compute_boundary_error", compute_boundary_error)
    setattr(CdsClass, "set_locus_tag_typo", set_locus_tag_typo)
    setattr(CdsClass, "compute_description_error", compute_description_error)
    setattr(CdsClass, "check_for_errors", check_for_errors)

# TODO refactor and test.
def set_start_end_strand_id(self):
    # Create a tuple of feature location data.
    # For start and end of feature, it doesn't matter whether
    # the feature is complex with a translational frameshift or not.
    # Retrieving the "start" and "end" attributes return the
    # very beginning and end of the feature,
    # disregarding the inner "join" coordinates.
    self._start_end_strand_id = (str(self.start),str(self.stop),self.orientation)

    # Since this id matched genes with different start sites,
    # the orientation impacts whether the left or right boundary is used
    if self.orientation == "forward":
        self._end_strand_id = (str(self.stop),self.orientation)
    elif self.orientation == "reverse":
        self._end_strand_id = (str(self.start),self.orientation)
    else:
        pass

# TODO refactor and test.
def set_search_genome_id(self):
    self._search_genome_id = basic.edit_suffix(self.genome_id, "remove").lower()


# TODO refactor and test.
def check_locus_tag(self):
    if self.locus_tag == "":
        self._locus_tag_missing = True

# TODO refactor and test.
def compute_amino_acid_errors(self, protein_alphabet):
    amino_acid_set = set(self.translation)
    amino_acid_error_set = amino_acid_set - protein_alphabet
    if len(amino_acid_error_set) > 0:
        self._amino_acid_errors = True

# TODO refactor and test.
def compute_boundary_error(self):
    # Check if start and end coordinates are fuzzy
    if not (str(self.start).isdigit() and str(self.stop).isdigit()):
        self._boundary_error = True

# TODO refactor and test.
def set_locus_tag_typo(self):
    self._locus_tag_typo = True

# TODO refactor and test.
def compute_description_error(self):
    # If the product description is empty or generic,
    # and the function or note descriptions are not, there is an error.
    if (self.product == "" and self.function != "" or self.note != ""):
        self._description_field_error = True

# TODO refactor and test.
def check_for_errors(self):
    if self._amino_acid_errors:
        self._total_errors += 1
    if self._boundary_error:
        self._total_errors += 1
    if self._description_field_error:
        self._total_errors += 1
    if self._locus_tag_missing:
        self._total_errors += 1
    if self._locus_tag_typo:
        self._total_errors += 1
    if self._unmatched_error:
        self._total_errors += 1




### Pipeline begins below

# TODO refactor and test.
def main(unparsed_args_list):
    """Run compare pipeline."""
    modify_genome_class(genome.Genome)
    modify_cds_class(cds.Cds)

    start_time = time.strftime("%x %X")
    args = parse_args(unparsed_args_list)
    database = args.database
    save_records = args.save_records
    interactive = args.interactive

    # Create config object with data obtained from file and/or defaults.
    config = configfile.build_complete_config(args.config_file)
    mysql_creds = config["mysql"]
    ncbi_creds = config["ncbi"]

    # Filters input: phage.Status=draft AND phage.HostGenus=Mycobacterium
    # Args structure: [['phage.Status=draft'], ['phage.HostGenus=Mycobacterium']]
    filters = args.filters

    # Setup output
    output_folder = basic.set_path(args.output_folder, kind="dir",
                                        expect=True)
    working_dir = pathlib.Path(WORKING_FOLDER)
    working_path = basic.make_new_dir(output_folder, working_dir,
                                      attempt=50)
    if working_path is None:
        print(f"Invalid working directory '{working_dir}'")
        sys.exit(1)

    results_path = pathlib.Path(working_path, RESULTS_FOLDER)
    error_path = pathlib.Path(working_path, ERROR_FOLDER)
    records_path = pathlib.Path(working_path, RECORD_FOLDER)
    results_path.mkdir()
    error_path.mkdir()
    records_path.mkdir()

    # Verify database connection and schema compatibility.
    print("Connecting to the MySQL database...")
    alchemist = AlchemyHandler(database=database,
                               username=mysql_creds["user"],
                               password=mysql_creds["password"])
    alchemist.connect(pipeline=True)
    alchemist.build_metadata()
    engine = alchemist.engine
    mysqldb.check_schema_compatibility(engine, "the compare pipeline")

    # Get PhageIDs, which also validates filters.
    phage_ids = get_ids(alchemist, filters, TARGET_TABLE)

    # Gather user-selected databases.
    valid_dbs = get_dbs(args.phagesdb, args.genbank)

    # Get version from the MySQL database.
    version = str(mysqldb_basic.scalar(engine, VERSION_QUERY))

    # Record the summary of settings for the comparison.
    record_compare_settings(working_path, database, version, filters, valid_dbs)

    # Now proceed with getting all genome data.
    pmd_tup = process_mysql_data(working_path, engine, phage_ids,
                                 interactive, save_records)
    mysql_genome_dict = pmd_tup[0]
    mysql_accessions = pmd_tup[1]
    mysql_acc_duplicates = pmd_tup[2]

    # Close connections since no need to interact with MySQL anymore.
    engine.dispose()

    #Now retrieve all PhagesDB data
    if "phagesdb" in valid_dbs:
        ppd_tup = process_phagesdb_data(working_path, interactive, save_records)
        pdb_genome_dict = ppd_tup[0]
        pdb_name_duplicates = ppd_tup[1]

    else:
        pdb_genome_dict = {}
        pdb_name_duplicates = set()

    # Retrieve and parse GenBank records if selected by user
    if "genbank" in valid_dbs:
        gbk_genome_dict = process_gbk_data(working_path, ncbi_creds,
                                mysql_accessions, interactive, save_records)
        # gbk_genome_dict: Key = accession; #Value = genome data
    else:
        gbk_genome_dict = {}

    # Now that all GenBank and PhagesDB data is retrieved,
    # match up to MySQL genome data.
    match_tup = match_all_genomes(mysql_genome_dict, pdb_genome_dict,
                      gbk_genome_dict, pdb_name_duplicates,
                      mysql_acc_duplicates)
    matched_genomes_list = match_tup[0]

    # Output unmatched data to file
    record_unmatched_pdb_data(match_tup[1], working_path)
    record_unmatched_gbk_data(match_tup[2], working_path)

    # Run checks on each genome and all matched data.
    check_mysql_gnms(mysql_genome_dict)
    check_pdb_gnms(pdb_genome_dict)
    check_gbk_gnms(gbk_genome_dict)
    check_matched_gnms(matched_genomes_list)

    # Now that all individual matched_genome_objects have all
    # computed attributes, compute database summary.
    summarize_data(matched_genomes_list, working_path)

    end_time = time.strftime("%x %X")
    print(f"Start time: {start_time}")
    print(f"Stop time: {end_time}")

    return


# TODO refactor and test.
def parse_args(unparsed_args_list):
    """Verify the correct arguments are selected for comparing databases."""

    compare_help = (
        "Pipeline to compare MySQL, PhagesDB, and "
        "GenBank databases for inconsistencies.")
    database_help = "Name of the MySQL database from which to compare data."
    output_folder_help = "Path to the folder to store results."
    phagesdb_help = "Indicates that PhagesDB data should be compared."
    genbank_help = "Indicates that GenBank data should be compared."
    save_records_help = (
        "Indicates that all genomes compared will be saved to file.")
    filters_help = (
        "Indicates which genomes to include/exclude from the comparison, "
        "with each conditional formatted as 'table.Field=value'.")
    interactive_help = (
        "Indicates whether evaluation is paused when errors are encountered.")
    config_file_help = "Path to the file containing user-specific login data."

    parser = argparse.ArgumentParser(description=compare_help)
    parser.add_argument("database", type=str, help=database_help)
    parser.add_argument("-o", "--output_folder", type=pathlib.Path,
                        default=pathlib.Path(DEFAULT_OUTPUT_FOLDER),
                        help=output_folder_help)
    parser.add_argument("-p", "--phagesdb", action="store_true",
        default=False, help=phagesdb_help)
    parser.add_argument("-g", "--genbank", action="store_true",
        default=False, help=genbank_help)
    parser.add_argument("-f", "--filters", nargs="?",
                        type=parsing.parse_cmd_string, help=filters_help,
                        default=[])
    parser.add_argument("-s", "--save_records", action="store_true",
        default=False, help=save_records_help)
    parser.add_argument("-i", "--interactive", action="store_true",
        default=False, help=interactive_help)
    parser.add_argument("-c", "--config_file", type=pathlib.Path,
                        help=config_file_help, default=None)


    # Assumed command line arg structure:
    # python3 -m pdm_utils <pipeline> <additional args...>
    # sys.argv:      [0]            [1]         [2...]
    args = parser.parse_args(unparsed_args_list[2:])
    return args



# TODO test.
def get_ids(alchemist, filters, target_table):
    """Get list of unique MySQL target table primary key values to evaluate."""

    primary_key = get_primary_key(alchemist.metadata, target_table)

    # Create filter object and then add command line filter strings
    db_filter = Filter(alchemist=alchemist, key=primary_key)
    db_filter.values = []

    # Attempt to add filters and exit if needed.
    add_filters(db_filter, filters)

    # Performs the query
    db_filter.update()

    # db_filter.values now contains list of PhageIDs that pass the filters.
    return db_filter.values


# TODO test.
def get_primary_key(metadata, target_table):
    """Get the primary key to the target table."""

    # Get SQLAlchemy metadata Table object
    # table_obj.primary_key.columns is a
    # SQLAlchemy ColumnCollection iterable object
    # Set primary key = 'phage.PhageID'
    table = querying.get_table(metadata, target_table)
    for column in table.primary_key.columns:
        primary_key = column
    return primary_key


# TODO test.
def add_filters(filter_obj, filters):
    """Add filters from command line to filter object."""
    errors = 0
    for or_filters in filters:
        for filter in or_filters:
            # Catch the error if it is an invalid table.column
            try:
                filter_obj.add(filter)
            except:
                print(f"Invalid filter: {filter}")
                errors += 1
    if errors > 0:
        print("Unable to run compare pipeline.")
        sys.exit(1)


# TODO refactor and test.
def get_dbs(pdb, gbk):
    """Create set of databases to compare to MySQL."""
    dbs = set()
    if pdb == True:
        dbs.add("phagesdb")
    if gbk == True:
        dbs.add("genbank")
    return dbs


# TODO refactor and test.
def output_to_file(data_list, folder, filename):
    """Output list data to file."""
    file_path = pathlib.Path(folder, filename)
    with file_path.open("w") as handle:
        writer = csv.writer(handle)
        for element in data_list:
            writer.writerow(element)


# TODO refactor and test.
def prepare_unmatched_to_gbk_output(gnms):
    """Prepare list of MySQL unmatched to GenBank data to be saved to file."""
    l = []
    if len(gnms) > 0:
        l.append(["PhageID", "PhageName", "Author", "Status", "Accession"])
        for gnm in gnms:
            l.append([gnm.id,
                      gnm.name,
                      gnm.annotation_author,
                      gnm.annotation_status,
                      gnm.accession])
    return l


# TODO refactor and test.
def prepare_unmatched_to_pdb_output(gnms):
    """Prepare list of MySQL unmatched to PhagesDB data to be saved to file."""
    l = []
    if len(gnms) > 0:
        l.append(["PhageID", "PhageName", "Author", "Status"])
        for gnm in gnms:
            l.append([gnm.id,
                      gnm.name,
                      gnm.annotation_author,
                      gnm.annotation_status])
    return l


# TODO refactor and test.
def save_seqrecord(seqrecord, output_path, file_prefix, ext,
                   seqrecord_ext, interactive):
    """Save record to file."""
    file_path = basic.make_new_file(output_path, file_prefix, ext, attempt=100)
    if file_path is not None:
        SeqIO.write(seqrecord, file_path, seqrecord_ext)
    else:
        print(f"Duplicated filenames for {file_prefix}. Unable to output data.")
        if interactive:
            input('Press ENTER to proceed')


# TODO refactor and test.
def record_compare_settings(working_path, database, version, filters, valid_dbs):
    """Save user-selected settings."""

    # Format setting strings
    lst = [
        ["Comparison date:", CURRENT_DATE],
        ["MySQL database:", database],
        ["MySQL database version:", version],
        ["Databases compared to MySQL:", ", ".join(valid_dbs)]
        ]

    # Add the filters
    count = 1
    for or_filters in filters:
        for filter in or_filters:
            lst.append([f"Filter {count}:", filter])
            count += 1

    output_to_file(lst, pathlib.Path(working_path), COMPARE_SETTINGS)

# TODO refactor and test.
def selected_authors_lst(lst1):
    lst1 = [str(i) for i in list(lst1)]
    lst2 =  ["Genomes with the following AnnotationAuthor will be compared:",
             ", ".join(lst1)]
    return lst2

# TODO refactor and test.
def process_mysql_data(working_path, engine, phage_ids, interactive, save):
    """Retrieve and process MySQL data."""

    print('\n\nPreparing genome data sets from the MySQL database...')
    genome_list = mysqldb.parse_genome_data(engine,
                                            phage_id_list=phage_ids,
                                            phage_query=PHAGE_QUERY,
                                            gene_query=GENE_QUERY,
                                            gnm_type=GNM_MYSQL)

    tup = filter_mysql_genomes(genome_list)
    gnm_dict = tup[0]
    name_duplicates = tup[2]
    accessions = tup[3]
    accession_dupes = tup[4]

    set_mysql_gnm_attr(gnm_dict)

    if save == True:
        print("Saving MySQL genomes to file...")
        save_gnms_to_fasta(gnm_dict, pathlib.Path(working_path, RECORD_FOLDER),
                           MYSQL_OUTPUT, interactive)

    # PhagesDB relies on the phageName, and not the phageID.
    # But the MySQL database does not require phageName values to be unique.
    # Check if there are any phageName duplications.
    # If there are, they will not be able to be compared to PhagesDB data.
    # Output duplicate phage search names to file
    if len(name_duplicates) > 0:
        print("Warning: There are duplicate phage names in the MySQL database.")
        print("Some MySQL genomes will not be able to be matched to PhagesDB.")
        lst1 = [[i] for i in list(name_duplicates)]
        output_to_file(lst1, pathlib.Path(working_path, ERROR_FOLDER),
                       DUPLICATE_MYSQL_NAMES)
        if interactive:
            input('Press ENTER to proceed')

    # Accessions aren't required to be unique in the MySQL
    # database (but they should be), so there could be duplicates
    # Output duplicate names to file
    if len(accession_dupes) > 0:
        print("Warning: There are duplicate accessions in the MySQL database.")
        print("Some MySQL genomes will not be able to be "
              "matched to GenBank records.")
        lst2 = [[i] for i in list(accession_dupes)]
        output_to_file(lst2, pathlib.Path(working_path, ERROR_FOLDER),
                       DUPLICATE_MYSQL_ACC)
        if interactive:
            input('Press ENTER to proceed')

    return gnm_dict, accessions, accession_dupes


# TODO refactor and test.
def filter_mysql_genomes(genome_list):
    """Only keep selected subset of genomes with no value duplications."""

    gnm_dict = {}
    names = set()
    duplicate_names = set()
    accessions = set()
    duplicate_accessions = set()

    for gnm in genome_list:

        gnm_dict[gnm.id] = gnm

        # This keeps track of whether there are duplicate phage
        # names that will be used to match up to PhagesDB data.
        if gnm.id in names:
            duplicate_names.add(gnm.id)
        else:
            names.add(gnm.id)

        # This keeps track of whether there are duplicate
        # accession numbers that will be used to match up to GenBank data.
        if gnm.accession != "":
            if gnm.accession in accessions:
                duplicate_accessions.add(gnm.accession)
            else:
                accessions.add(gnm.accession)

    return (gnm_dict, names, duplicate_names, accessions, duplicate_accessions)


# TODO refactor and test.
def set_mysql_gnm_attr(gnm_dict):
    """Set compare-specific attributes not in pdm_utils Genome class."""
    for key in gnm_dict.keys():
        gnm = gnm_dict[key]
        for cds_ftr in gnm.cds_features:
            set_mysql_cds_attr(cds_ftr)

# TODO refactor and test.
def save_gnms_to_fasta(gnm_dict, main_path, new_dir, interactive):
    """Save genome data to fasta file."""

    output_path = pathlib.Path(main_path, new_dir)
    output_path.mkdir()

    for key in gnm_dict.keys():
        gnm = gnm_dict[key]
        fasta = SeqRecord(gnm.seq, id=gnm.id, description="")
        save_seqrecord(fasta, output_path, gnm.id, "fasta",
                       "fasta", interactive)


# TODO refactor and test.
def set_mysql_cds_attr(cds_ftr):
    """Set compare-specific MySQL Cds attributes not in pdm_utils Cds class."""
    cds_ftr.set_search_genome_id()
    cds_ftr.set_start_end_strand_id()
    cds_ftr.type = CDS_MYSQL

# TODO refactor and test.
def check_mysql_cds(cds_ftr):
    """Check for errors in MySQL CDS feature."""
    cds_ftr.compute_amino_acid_errors(constants.PROTEIN_ALPHABET)
    cds_ftr.compute_boundary_error()


# TODO refactor and test.
def check_mysql_gnms(gnm_dict):
    """Check for errors in MySQL matched genome and CDS features."""
    for id in gnm_dict.keys():
        gnm = gnm_dict[id]

        for cds_ftr in gnm.cds_features:
            check_mysql_cds(cds_ftr)

        gnm.check_status_accession()
        gnm.compute_nucleotide_errors(constants.DNA_ALPHABET)
        gnm.compute_cds_feature_errors()
        gnm.compute_status_description_error()


# TODO refactor and test.
def process_phagesdb_data(working_path, interactive, save):
    """Retrieve data from PhagesDB and process results."""

    gbd_tup = get_pdb_data(interactive)
    gnm_dict = gbd_tup[0]
    names = gbd_tup[1]
    name_dupes = gbd_tup[2]

    if save == True:
        print("Saving PhagesDB genomes to file...")
        save_gnms_to_fasta(gnm_dict, pathlib.Path(working_path, RECORD_FOLDER),
                           PHAGESDB_OUTPUT, interactive)

    # PhagesDB phage names should be unique, but just make sure after
    # they are converted to a search name.
    if len(name_dupes) > 0:

        print("Warning: There are duplicate phage search names in PhagesDB.")
        print("Some PhagesDB genomes will not be able to be "
              "matched to genomes in MySQL.")
        lst = [[i] for i in list(name_dupes)]
        output_to_file(lst, pathlib.Path(working_path, ERROR_FOLDER),
                       DUPLICATE_PDB_NAMES)
        if interactive:
            input('Press ENTER to proceed')

    return gnm_dict, name_dupes


# TODO refactor and test.
def get_pdb_data(interactive):
    """Retrieve data from PhagesDB."""
    print('\n\nRetrieving data from PhagesDB...')
    gnm_dict = {}
    names = set()
    dupe_names = set()
    data_list = phagesdb.get_phagesdb_data(constants.API_SEQUENCED)

    if len(data_list) > 0:
        for i in range(len(data_list)):
            # Note: this step take a long time because it needs to retrieve
            # and parse fasta files to get the genome sequence.
            gnm = phagesdb.parse_genome_data(data_list[i], gnm_type=GNM_PDB,
                                             seq=True)
            pdb_id = gnm.id
            if pdb_id in names:
                dupe_names.add(pdb_id)
            else:
                names.add(pdb_id)
                gnm_dict[pdb_id] = gnm
    else:
        print("Error retrieving PhagesDB data.")
        if interactive:
            input('Press ENTER to proceed')
    return gnm_dict, names, dupe_names


# TODO refactor and test.
def check_pdb_gnms(gnm_dict):
    """Check for errors in PhagesDB genome."""
    for id in gnm_dict.keys():
        gnm = gnm_dict[id]
        if str(gnm.seq) != "":
            gnm.compute_nucleotide_errors(constants.DNA_ALPHABET)

# TODO refactor and test.
def process_gbk_data(working_path, ncbi_creds, accessions, interactive, save):
    """Retrieve and process GenBank data."""

    if save == True:
        output_path = pathlib.Path(working_path, RECORD_FOLDER, GENBANK_OUTPUT)
        output_path.mkdir()

    records, retrieval_errors = get_genbank_data(ncbi_creds, accessions)

    #Report the accessions that could not be retrieved.
    output_to_file(retrieval_errors, pathlib.Path(working_path, ERROR_FOLDER),
                   FAILED_ACC_RETRIEVE)

    gnm_dict = {}
    for record in records:
        gnm = flat_files.parse_genome_data(record,gnm_type=GNM_GBK)
        set_gbk_gnm_attr(gnm)
        gnm_dict[gnm.accession] = gnm
        if save == True:
            save_gbk_genome(gnm, record, output_path, interactive)

    return gnm_dict


# TODO refactor and test.
def get_genbank_data(ncbi_cred_dict, accession_set, batch_size=200):
    """Retrieve genomes from GenBank."""

    print("\n\nRetrieving records from GenBank")

    # Use esearch to verify the accessions are valid and
    # efetch to retrieve the record.
    ncbi.set_entrez_credentials(
        tool=ncbi_cred_dict["tool"],
        email=ncbi_cred_dict["email"],
        api_key=ncbi_cred_dict["api_key"])


    #Create batches of accessions
    acc_list = list(accession_set)

    #Add [ACCN] field to each accession number
    for index in range(len(acc_list)):
        acc_list[index] = acc_list[index] + "[ACCN]"

    #Keep track of specific records
    retrieved_records = []
    retrieval_errors = []

    batch_indices = basic.create_indices(acc_list, batch_size)
    for indices in batch_indices:
        start = indices[0]
        stop = indices[1]

        current_batch_size = stop - start
        delimiter = " | "
        esearch_term = delimiter.join(acc_list[start:stop])

        #Use esearch for each accession
        search_handle = Entrez.esearch(db="nucleotide", term=esearch_term,
                                       usehistory="y")
        search_record = Entrez.read(search_handle)
        search_count = int(search_record["Count"])
        search_webenv = search_record["WebEnv"]
        search_query_key = search_record["QueryKey"]

        # Keep track of the accessions that failed to be located in GenBank
        if search_count < current_batch_size:
            search_acc_fail = search_record["ErrorList"]["PhraseNotFound"]

            # Each element in this list is formatted "accession[ACCN]"
            for element in search_acc_fail:
                retrieval_errors.append(element[:-6])

        # Now retrieve all records using efetch
        fetch_handle = Entrez.efetch(db="nucleotide",
                                    rettype="gb",
                                    retmode="text",
                                    retstart=0,
                                    retmax=search_count,
                                    webenv=search_webenv,
                                    query_key=search_query_key)
        fetch_records = SeqIO.parse(fetch_handle,"genbank")

        for record in fetch_records:
            retrieved_records.append(record)

        search_handle.close()
        fetch_handle.close()
    return retrieved_records, retrieval_errors


# TODO refactor and test.
def set_gbk_gnm_attr(gnm):
    """Set compare-specific attributes not in pdm_utils Genome class."""
    gnm.record_name = ""
    gnm.record_id = ""

    if len(gnm.source_features) == 1:
        src_ftr = gnm.source_features[0]
        gnm.source_feature_organism = src_ftr.organism
        gnm.source_feature_host = src_ftr.host
        gnm.source_feature_lab_host = src_ftr.lab_host

    for cds_ftr in gnm.cds_features:
        set_gbk_cds_attr(cds_ftr)


# TODO refactor and test.
def check_gbk_gnms(gnm_dict):
    """Check for errors in GenBank genome and CDS features."""
    for id in gnm_dict.keys():
        gnm = gnm_dict[id]

        for cds_ftr in gnm.cds_features:
            check_gbk_cds(cds_ftr)
        gnm.compute_nucleotide_errors(constants.DNA_ALPHABET)
        gnm.compute_cds_feature_errors()
        gnm.compute_gbk_cds_feature_errors()


# TODO refactor and test.
def set_gbk_cds_attr(cds_ftr):
    """Set compare-specific Cds attributes not in pdm_utils Cds class."""
    cds_ftr.set_start_end_strand_id()
    cds_ftr.type = CDS_GBK


# TODO refactor and test.
def check_gbk_cds(cds_ftr):
    """Check for errors in GenBank CDS feature."""
    cds_ftr.check_locus_tag()
    cds_ftr.compute_amino_acid_errors(constants.PROTEIN_ALPHABET)
    cds_ftr.compute_boundary_error()
    cds_ftr.compute_description_error()


# TODO refactor and test.
def save_gbk_genome(gnm, record, output_path, interactive):
    """Save GenBank record to file."""
    file_prefix = gnm.id + "__" + gnm.accession
    save_seqrecord(record, output_path, file_prefix,
                   "gb", "genbank", interactive)


# TODO refactor and test.
def match_all_genomes(mysql_gnms, pdb_gnms, gbk_gnms, pdb_name_duplicates,
                      mysql_acc_duplicates):
    """Match MySQL, PhagesDB, and GenBank genomes."""

    print("Matching PhagesDB and GenBank genomes to MySQL genomes...")
    matched_gnms = []
    mysql_unmatched_to_pdb_gnms = []
    mysql_unmatched_to_gbk_gnms = []
    match_count = 1
    total_count = len(mysql_gnms.keys())
    # Iterate through each phage in the MySQL database
    for id in mysql_gnms.keys():

        print("Matching genome %s of %s" %(match_count,total_count))
        mysql_gnm = mysql_gnms[id]
        gnm_triad = genometriad.GenomeTriad()
        gnm_triad.m_genome = mysql_gnm

        # Match up PhagesDB genome
        # First try to match up the phageID, and if that doesn't work,
        # try to match up the phageName
        if mysql_gnm.id in pdb_gnms.keys():
            pdb_genome = pdb_gnms[mysql_gnm.id]
            # Make sure the pdb_genome doesn't have a search name
            # that was duplicated
            if pdb_genome.id in pdb_name_duplicates:
                # Do NOT set genome.type, so that it remains ""
                pdb_genome = genome.Genome()

        else:
            # Do NOT set genome.type, so that it remains ""
            pdb_genome = genome.Genome()
            mysql_unmatched_to_pdb_gnms.append(mysql_gnm)

        gnm_triad.p_genome = pdb_genome

        # Now match up GenBank genome
        acc = mysql_gnm.accession
        if acc != "" and acc not in mysql_acc_duplicates:
                # Retrieval of record may have failed
                try:
                    g_gnm = gbk_gnms[acc]
                except:
                    # Do NOT set genome.type, so that it remains ""
                    g_gnm = genome.Genome()
                    mysql_unmatched_to_gbk_gnms.append(mysql_gnm)

        else:
            # Do NOT set genome.type, so that it remains ""
            g_gnm = genome.Genome()
            mysql_unmatched_to_gbk_gnms.append(mysql_gnm)

        gnm_triad.g_genome = g_gnm
        matched_gnms.append(gnm_triad)
        match_count += 1

    return (matched_gnms, mysql_unmatched_to_pdb_gnms, mysql_unmatched_to_gbk_gnms)

# TODO refactor and test.
def record_unmatched_gbk_data(gnms, working_path):
    """Save unmatched MySQL data not matched to GenBank."""
    lst1 = prepare_unmatched_to_gbk_output(gnms)
    output_to_file(lst1, pathlib.Path(working_path, ERROR_FOLDER),
                   FAILED_ACC_RETRIEVE)

# TODO refactor and test.
def record_unmatched_pdb_data(gnms, working_path):
    """Save unmatched MySQL data not matched to PhagesDB."""
    lst2 = prepare_unmatched_to_pdb_output(gnms)
    output_to_file(lst2, pathlib.Path(working_path, ERROR_FOLDER),
                   UNMATCHED_GENOMES)

# TODO refactor and test.
def check_matched_gnms(gnm_triads):
    """Compare all matched data."""

    print("Comparing matched genomes and identifying inconsistencies...")
    count = 1
    total = len(gnm_triads)
    for gnm_triad in gnm_triads:
        print(f"Comparing matched genome set {count} of {total}")

        gnm_triad.compare_mysql_gbk_genomes(GNM_MYSQL, GNM_GBK,
                                            CDSPAIR_MYSQL_GBK)
        gnm_triad.compare_mysql_phagesdb_genomes(GNM_MYSQL, GNM_PDB)
        gnm_triad.compare_phagesdb_gbk_genomes(GNM_PDB, GNM_GBK)
        gnm_triad.compute_total_genome_errors(GNM_GBK, GNM_PDB)
        count += 1


# TODO refactor and test.
def summarize_data(matched_genomes_list, working_path):
    """Create summary of data and save."""
    summary = dbcomparesummary.DbCompareSummary(matched_genomes_list)
    summary.compute_summary(GNM_MYSQL, GNM_PDB, GNM_GBK)
    output_all_data(working_path, summary)


# TODO refactor and test.
def output_all_data(output_path, summary):
    """Output all analysis results."""
    print("Outputting results to file...")

    cds_file_path = pathlib.Path(output_path, RESULTS_FOLDER, GENE_OUTPUT)
    cds_fh = cds_file_path.open("w")
    cds_writer = csv.writer(cds_fh)
    cds_headers = create_gene_headers()
    cds_writer.writerow(cds_headers)

    #Create genome summary output file
    lst1 = create_genome_summary_data(summary)
    output_to_file(lst1, pathlib.Path(output_path, RESULTS_FOLDER),
                   GNM_SUMMARY_OUTPUT)

    #Create CDS summary output file
    lst3 = create_cds_summary_data(summary)
    output_to_file(lst3, pathlib.Path(output_path, RESULTS_FOLDER),
                   CDS_SUMMARY_OUTPUT)

    # Now iterate through matched objects.
    # All MySQL genomes are stored in a GenomeTriad object,
    # even if there are no PhagesDB or GenBank matches.
    # All but a few MySQL phages should be matched to PhagesDB
    # Many MySQL phages should be matched to GenBank
    genomes_data = []
    for gnm_triad in summary._matched_genomes_list:

        gnm_data = create_genome_data(gnm_triad)
        genomes_data.append(gnm_data)

        #Once all matched genome data has been outputted,
        # iterate through all matched gene data
        m_gnm = gnm_triad.m_genome
        all_ftrs = get_all_features(gnm_triad)
        for mixed_ftr in all_ftrs:
            ftr_data = create_feature_data(m_gnm, mixed_ftr)
            cds_writer.writerow(ftr_data)

    lst2 = [create_genome_headers()]
    lst2.extend(genomes_data)
    output_to_file(lst2, pathlib.Path(output_path, RESULTS_FOLDER),
                   GENOME_OUTPUT)
    cds_fh.close()
    return


# TODO refactor and test.
def get_all_features(gnm_triad):
    """Create list of all features."""

    perfect_match = gnm_triad._m_g_perfect_matched_ftrs
    imperfectly_match = gnm_triad._m_g_imperfect_matched_ftrs
    mysql_unmatch = gnm_triad._m_ftrs_unmatched_in_g
    gbk_unmatch = gnm_triad._g_ftrs_unmatched_in_m

    lst = []
    lst.extend(perfect_match)
    lst.extend(imperfectly_match)
    lst.extend(mysql_unmatch)
    lst.extend(gbk_unmatch)

    return lst


# TODO refactor and test.
def create_gene_headers():
    """Create list of column headers."""
    headers = [
        # Gene summary
        "mysql_phage_name",
        "total_errors",

        # MySQL
        # General gene data
        "mysql_phage_id",
        "mysql_phage_id",
        "mysql_type_id",
        "mysql_gene_id",
        "mysql_gene_name",
        "mysql_left_boundary",
        "mysql_right_boundary",
        "mysql_strand",
        "mysql_translation",
        "mysql_translation_length",
        "mysql_gene_notes",

        # Gene data checks
        "mysql_translation_error",
        "mysql_gene_coords_error",

        # GenBank
        # General gene data
        "gbk_locus_tag",
        "gbk_gene_number",
        "gbk_type_id",
        "gbk_left_boundary",
        "gbk_right_boundary",
        "gbk_strand",
        "gbk_translation",
        "gbk_translation_length",
        "gbk_product_description",
        "gbk_function_description",
        "gbk_note_description",

        # Gene data checks
        "gbk_translation_error",
        "gbk_gene_coords_error",
        "gbk_missing_locus_tag",
        "gbk_locus_tag_typo",
        "gbk_description_field_error",

        # MySQL-GenBank checks
        "mysql_gbk_unmatched_error",
        "mysql_gbk_description_error",
        "mysql_gbk_start_coordinate_error",
        "mysql_gbk_translation_error"
        ]
    return headers


# TODO refactor and test.
def create_genome_headers():
    """Create list of column headers."""
    headers = [
        "mysql_phage_id",
        "contains_errors",

        # MySQL
        # General genome data
        "mysql_phage_name",
        "mysql_phage_id",
        "mysql_phage_name",
        "mysql_status",
        "mysql_cluster",
        "mysql_host",
        "mysql_accession",
        "mysql_dna_seq_length",
        "mysql_gene_tally",
        "mysql_description_tally",
        "mysql_gbk_update_flag",
        "mysql_date_last_modified",
        "mysql_annotation_author",

        # Genome data checks
        "mysql_dna_seq_error",
        "mysql_gene_translation_error_tally",
        "mysql_gene_coords_error_tally",
        "mysql_status_description_error",
        "mysql_status_accession_error",

        # PhagesDB
        # General genome data
        "pdb_phage_name",
        "pdb_phage_name",
        "pdb_cluster",
        "pdb_subcluster",
        "pdb_host",
        "pdb_accession",
        "pdb_dna_seq_length",

        # Genome data checks
        "pdb_dna_seq_error",

        # GenBank
        # General genome data
        "gbk_phage_name",
        "gbk_phage_name",
        "gbk_record_id",
        "gbk_record_name",
        "gbk_record_accession",
        "gbk_record_definition",
        "gbk_record_source",
        "gbk_record_organism",
        "gbk_source_feature_organism",
        "gbk_source_feature_host",
        "gbk_source_feature_lab_host",
        "gbk_authors",
        "gbk_dna_seq_length",
        "gbk_gene_tally",

        # Genome data checks
        "gbk_dna_seq_error",
        "gbk_gene_translation_error_tally",
        "gbk_gene_coords_error_tally",
        "gbk_gene_product_tally",
        "gbk_gene_function_tally",
        "gbk_gene_note_tally",
        "gbk_missing_locus_tag_tally",
        "gbk_locus_tag_typo_tally",
        "gbk_description_field_error_tally",

        # MySQL-PhagesDB
        "mysql_pdb_dna_seq_error",
        "mysql_pdb_dna_seq_length_error",
        "mysql_pdb_cluster_error",
        "mysql_pdb_accession_error",
        "mysql_pdb_host_error",

        # MySQL-GenBank
        "mysql_gbk_dna_seq_error",
        "mysql_gbk_dna_seq_length_error",
        "mysql_gbk_record_header_name_error",
        "mysql_gbk_record_header_host_error",

        # Author error is dependent on MySQL genome annotation author and
        # GenBank list of authors, so this metric should be reported with
        # the other mysql_gbk error tallies.
        "mysql_gbk_author_error",
        "mysql_gbk_perfectly_matched_gene_tally",
        "mysql_gbk_imperfectly_matched_gene_tally",
        "mysql_gbk_unmatched_mysql_gene_tally",
        "mysql_gbk_unmatched_gbk_gene_tally",
        "mysql_gbk_gene_description_error_tally",
        "mysql_gbk_perfectly_matched_gene_translation_error_tally",

        # Number of genes with errors is computed slightly differently
        # depending on whethere there are matching MySQL and GenBank genomes.
        # Therefore,this metric should be reported with the other mysql_gbk
        #  error tallies even if there is no matching GenBank genome.
        "mysql_gbk_genes_with_errors_tally",

        # PhagesDB-GenBank
        "pdb_gbk_dna_seq_error",
        "pdb_gbk_dna_seq_length_error"
        ]
    return headers


# TODO refactor and test.
def create_genome_summary_fields():
    """Create genome summary row ids."""
    lst = [
        # Column header
        "Database comparison metric",

        # Database summaries
        "mysql_total_genomes_analyzed",
        "mysql_genomes_unmatched_to_pdb_tally",
        "mysql_genomes_unmatched_to_gbk_tally",
        "total_genomes_with_errors",


        # MySQL data
        # General genome data
        "mysql_gbk_update_flag_tally",

        # Genome data checks
        "mysql_genomes_with_nucleotide_errors_tally",
        "mysql_genomes_with_translation_errors_tally",
        "mysql_genomes_with_boundary_errors_tally",
        "mysql_genomes_with_status_accession_error_tally",
        "mysql_genomes_with_status_description_error_tally",

        # PhagesDB data
        # Genome data checks
        "pdb_genomes_with_nucleotide_errors_tally",

        # GenBank data
        # Genome data checks
        "gbk_genomes_with_description_field_errors_tally",
        "gbk_genomes_with_nucleotide_errors_tally",
        "gbk_genomes_with_translation_errors_tally",
        "gbk_genomes_with_boundary_errors_tally",
        "gbk_genomes_with_missing_locus_tags_tally",
        "gbk_genomes_with_locus_tag_typos_tally",

        # MySQL-PhagesDB checks
        "mysql_pdb_sequence_mismatch_tally",
        "mysql_pdb_sequence_length_mismatch_tally",
        "mysql_pdb_cluster_mismatch_tally",
        "mysql_pdb_accession_mismatch_tally",
        "mysql_pdb_host_mismatch_tally",

        # MySQL-GenBank checks
        "mysql_gbk_sequence_mismatch_tally",
        "mysql_gbk_sequence_length_mismatch_tally",
        "mysql_gbk_record_header_phage_mismatch_tally",
        "mysql_gbk_record_header_host_mismatch_tally",
        "mysql_gbk_genomes_with_author_errors_tally",
        "mysql_gbk_genomes_with_imperfectly_matched_features_tally",
        "mysql_gbk_genomes_with_unmatched_mysql_features_tally",
        "mysql_gbk_genomes_with_unmatched_gbk_features_tally",
        "mysql_gbk_genomes_with_different_descriptions_tally",
        "mysql_gbk_genomes_with_different_translations_tally",

        # PhagesDB-GenBank checks
        "pdb_gbk_sequence_mismatch_tally",
        "pdb_gbk_sequence_length_mismatch_tally"
        ]
    return lst


# TODO refactor and test.
def create_cds_summary_fields():
    """Create CDS summary row ids."""
    lst = [
        # Column header
        "Database comparison metric",

        # MySQL feature
        # Gene data checks
        "mysql__translation_errors_tally",
        "mysql__boundary_errors_tally",

        # GenBank feature
        # Gene data checks
        "gbk_translation_errors_tally",
        "gbk_boundary_errors_tally",
        "gbk_missing_locus_tags_tally",
        "gbk_locus_tag_typos_tally",
        "gbk_description_field_errors_tally",

        # MySQL-GenBank checks
        "mysql_gbk_different_descriptions_tally",
        "mysql_gbk_different_start_sites_tally",
        "mysql_gbk_different_translation_tally",
        "mysql_gbk_unmatched_mysql_features_tally",
        "mysql_gbk_unmatched_gbk_features_tally"
        ]
    return lst


# TODO refactor and test.
def create_genome_summary_data(summary):
    """Create summary of all genome results."""

    lst1 = [
        # Column header
        "tally",

        # First output database summary data
        summary._m_total_gnms_analyzed,
        summary._m_gnms_unmatched_to_p_tally,
        summary._m_gnms_unmatched_to_g_tally,
        summary._total_genomes_with_errors,

        # MySQL data
        # General genome data
        summary._m_g_update_flag_tally,

        # Genome data checks
        summary._m_gnms_with_nucleotide_errors_tally,
        summary._m_gnms_with_translation_errors_tally,
        summary._m_gnms_with_boundary_errors_tally,
        summary._m_gnms_with_status_accession_error_tally,
        summary._m_gnms_with_status_description_error_tally,

        # PhagesDB data
        # Genome data checks
        summary._p_gnms_with_nucleotide_errors_tally,

        # GenBank data
        # Genome data checks
        summary._g_gnms_with_description_field_errors_tally,
        summary._g_gnms_with_nucleotide_errors_tally,
        summary._g_gnms_with_translation_errors_tally,
        summary._g_gnms_with_boundary_errors_tally,
        summary._g_gnms_with_missing_locus_tags_tally,
        summary._g_gnms_with_locus_tag_typos_tally,


        # MySQL-PhagesDB checks
        summary._m_p_seq_mismatch_tally,
        summary._m_p_seq_length_mismatch_tally,
        summary._m_p_cluster_mismatch_tally,
        summary._m_p_accession_mismatch_tally,
        summary._m_p_host_mismatch_tally,

        # MySQL-GenBank checks
        summary._m_g_seq_mismatch_tally,
        summary._m_g_seq_length_mismatch_tally,
        summary._m_g_header_phage_mismatch_tally,
        summary._m_g_header_host_mismatch_tally,
        summary._m_g_gnms_with_author_errors_tally,
        summary._m_g_gnms_with_imperfectly_matched_ftrs_tally,
        summary._m_g_gnms_with_unmatched_m_ftrs_tally,
        summary._m_g_gnms_with_unmatched_g_ftrs_tally,
        summary._m_g_gnms_with_different_descriptions_tally,
        summary._m_g_gnms_with_different_translations_tally,

        # PhagesDB-GenBank checks
        summary._p_g_seq_mismatch_tally,
        summary._p_g_seq_length_mismatch_tally
        ]

    lst2 = create_genome_summary_fields()
    lst3 = []
    for i in range(len(lst2)):
        lst3.append([lst2[i],lst1[i]])
    return lst3


# TODO refactor and test.
def create_cds_summary_data(summary):
    """Create summary of all CDS results."""
    lst1 = [
        # Column header
        "tally",

        # MySQL feature
        # Gene data checks
        summary._m_translation_errors_tally,
        summary._m_boundary_errors_tally,

        # GenBank feature
        # Gene data checks
        summary._g_translation_errors_tally,
        summary._g_boundary_errors_tally,
        summary._g_missing_locus_tags_tally,
        summary._g_locus_tag_typos_tally,
        summary._g_description_field_errors_tally,

        # MySQL-GenBank checks
        summary._m_g_different_descriptions_tally,
        summary._m_g_different_start_sites_tally,
        summary._m_g_different_translation_tally,
        summary._m_g_unmatched_m_ftrs_tally,
        summary._m_g_unmatched_g_ftrs_tally
        ]

    lst2 = create_cds_summary_fields()
    lst3 = []
    for i in range(len(lst2)):
        lst3.append([lst2[i],lst1[i]])
    return lst3

# TODO refactor and test.
def create_feature_data(mysql_gnm, mixed_ftr):
    """Create feature data to output."""

    lst = []

    # Gene summaries
    # Add MySQL genome search name to each gene row regardless of
    # the type of CDS data (matched or unmatched).
    lst.append(mysql_gnm.id) # matched MySQL search name
    lst.append(mixed_ftr._total_errors) # total # of errors for this gene

    # Default type attribute is empty ""
    cds_pair = cdspair.CdsPair()
    mysql_ftr = cds.Cds()
    gbk_ftr = cds.Cds()

    # Determine what exactly the mixed feature object is and assign to
    # the appropriate variable.
    if isinstance(mixed_ftr, cdspair.CdsPair):
        cds_pair = mixed_ftr
        mysql_ftr = mixed_ftr.cds1
        gbk_ftr = mixed_ftr.cds2
    elif isinstance(mixed_ftr, cds.Cds):
        if mixed_ftr.type == CDS_MYSQL:
            mysql_ftr = mixed_ftr
        elif mixed_ftr.type == CDS_GBK:
            gbk_ftr = mixed_ftr
        else:
            pass
    else:
        pass


    # MySQL feature
    if mysql_ftr.type == CDS_MYSQL:

        # General gene data
        lst.append(mysql_ftr.genome_id)
        lst.append(mysql_ftr._search_genome_id)
        lst.append(mysql_ftr.type)
        lst.append(mysql_ftr.id)
        lst.append(mysql_ftr.name)
        lst.append(mysql_ftr.start)
        lst.append(mysql_ftr.stop)
        lst.append(mysql_ftr.orientation)
        lst.append(mysql_ftr.translation)
        lst.append(mysql_ftr.translation_length)
        lst.append(mysql_ftr.raw_description)

        # Gene data checks
        lst.append(mysql_ftr._amino_acid_errors)
        lst.append(mysql_ftr._boundary_error)

    else:
        lst.extend(["","","","","","","","","","","","",""])


    # GenBank feature
    if gbk_ftr.type == CDS_GBK:

        # General gene data
        lst.append(gbk_ftr.locus_tag)
        lst.append(gbk_ftr.gene)
        lst.append(gbk_ftr.type)
        lst.append(gbk_ftr.start)
        lst.append(gbk_ftr.stop)
        lst.append(gbk_ftr.orientation)
        lst.append(gbk_ftr.translation)
        lst.append(gbk_ftr.translation_length)
        lst.append(gbk_ftr.raw_product)
        lst.append(gbk_ftr.raw_function)
        lst.append(gbk_ftr.raw_note)

        # Gene data checks
        lst.append(gbk_ftr._amino_acid_errors)
        lst.append(gbk_ftr._boundary_error)
        lst.append(gbk_ftr._locus_tag_missing)
        lst.append(gbk_ftr._locus_tag_typo)
        lst.append(gbk_ftr._description_field_error)
    else:
        lst.extend(["","","","","","","","","","","","","","","",""])


    # MySQL-GenBank checks
    if cds_pair.type == CDSPAIR_MYSQL_GBK:

        # If this is a matched CDS feature, both MySQL and
        # GenBank features should have identical unmatched_error value.
        lst.append(cds_pair.cds1._unmatched_error)

        lst.append(cds_pair.different_description)
        lst.append(cds_pair.different_start_site)
        lst.append(cds_pair.different_translation)
    else:
        lst.append(mixed_ftr._unmatched_error)
        lst.extend(["","",""])
    return lst



# TODO refactor and test.
def create_genome_data(gnm_triad):
    """Create genome data to output."""

    mysql_gnm = gnm_triad.m_genome
    pdb_gnm = gnm_triad.p_genome
    gbk_gnm = gnm_triad.g_genome

    lst = []

    # Genome summary data
    lst.append(mysql_gnm.id)
    lst.append(gnm_triad._contains_errors)

    # MySQL data
    # General genome data
    lst.append(mysql_gnm.name)
    lst.append(mysql_gnm.id)
    lst.append(mysql_gnm.name)
    lst.append(mysql_gnm.annotation_status)
    lst.append(mysql_gnm.cluster)
    lst.append(mysql_gnm.host_genus)
    lst.append(mysql_gnm.accession)
    lst.append(mysql_gnm.length)
    lst.append(mysql_gnm._cds_features_tally)
    lst.append(mysql_gnm._cds_descriptions_tally)
    lst.append(mysql_gnm.retrieve_record)
    lst.append(mysql_gnm.date)
    lst.append(mysql_gnm.annotation_author)


    # Genome data checks
    lst.append(mysql_gnm._nucleotide_errors)
    lst.append(mysql_gnm._cds_features_with_translation_error_tally)
    lst.append(mysql_gnm._cds_features_boundary_error_tally)
    lst.append(mysql_gnm._status_description_error)
    lst.append(mysql_gnm._status_accession_error)


    # PhagesDB data
    if pdb_gnm.type == GNM_PDB:

        # General genome data
        lst.append(pdb_gnm.name)
        lst.append(pdb_gnm.name)
        lst.append(pdb_gnm.cluster)
        lst.append(pdb_gnm.subcluster)
        lst.append(pdb_gnm.host_genus)
        lst.append(pdb_gnm.accession)
        lst.append(pdb_gnm.length)

        # Genome data checks
        lst.append(pdb_gnm._nucleotide_errors)
    else:
        lst.extend(["","","","","","","",""])



    # GenBank data
    if gbk_gnm.type == GNM_GBK:

        # General genome data
        lst.append(gbk_gnm.name)
        lst.append(gbk_gnm.name)
        lst.append(gbk_gnm.record_id)
        lst.append(gbk_gnm.record_name)
        lst.append(gbk_gnm.accession)
        lst.append(gbk_gnm.description)
        lst.append(gbk_gnm.source)
        lst.append(gbk_gnm.organism)
        lst.append(gbk_gnm.source_feature_organism)
        lst.append(gbk_gnm.source_feature_host)
        lst.append(gbk_gnm.source_feature_lab_host)
        lst.append(gbk_gnm.authors)
        lst.append(gbk_gnm.length)
        lst.append(gbk_gnm._cds_features_tally)


        # Genome data checks
        lst.append(gbk_gnm._nucleotide_errors)
        lst.append(gbk_gnm._cds_features_with_translation_error_tally)
        lst.append(gbk_gnm._cds_features_boundary_error_tally)
        lst.append(gbk_gnm._cds_products_tally)
        lst.append(gbk_gnm._cds_functions_tally)
        lst.append(gbk_gnm._cds_notes_tally)
        lst.append(gbk_gnm._missing_locus_tags_tally)
        lst.append(gbk_gnm._locus_tag_typos_tally)
        lst.append(gbk_gnm._description_field_error_tally)

    else:
        lst.extend(["","","","","","","","","","",
                    "","","","","","","","","","",
                    "","",""])

    # MySQL-PhagesDB checks
    if pdb_gnm.type == GNM_PDB:

        lst.append(gnm_triad._m_p_seq_mismatch)
        lst.append(gnm_triad._m_p_seq_length_mismatch)
        lst.append(gnm_triad._m_p_cluster_mismatch)
        lst.append(gnm_triad._m_p_accession_mismatch)
        lst.append(gnm_triad._m_p_host_mismatch)
    else:
        lst.extend(["","","","",""])


    # MySQL-GenBank checks
    if gbk_gnm.type == GNM_GBK:
        lst.append(gnm_triad._m_g_seq_mismatch)
        lst.append(gnm_triad._m_g_seq_length_mismatch)
        lst.append(gnm_triad._g_header_fields_name_mismatch)
        lst.append(gnm_triad._g_host_mismatch)
        lst.append(gnm_triad._m_g_author_error)
        lst.append(gnm_triad._m_g_perfect_matched_ftrs_tally)
        lst.append(gnm_triad._m_g_imperfect_matched_ftrs_tally)
        lst.append(gnm_triad._m_ftrs_unmatched_in_g_tally)
        lst.append(gnm_triad._g_ftrs_unmatched_in_m_tally)
        lst.append(gnm_triad._m_g_different_descriptions_tally)
        lst.append(gnm_triad._m_g_different_translations_tally)
    else:
        lst.extend(["","","","","","","","","","",""])

    # Number of genes with errors
    lst.append(gnm_triad._total_number_genes_with_errors)

    # Output PhagesDB-GenBank checks
    if pdb_gnm.type == GNM_PDB and gbk_gnm.type == GNM_GBK:
        lst.append(gnm_triad._p_g_seq_mismatch)
        lst.append(gnm_triad._p_g_seq_length_mismatch)
    else:
        lst.extend(["",""])
    return lst



"""Represents a collection of data about a tRNA feature that are commonly used
to maintain and update SEA-PHAGES phage genomics data.
"""

import re
from collections import OrderedDict

from Bio.Alphabet import IUPAC
from Bio.Seq import Seq
from Bio.SeqFeature import SeqFeature, FeatureLocation, CompoundLocation
from pdm_utils.classes import evaluation
from pdm_utils.classes.aragornhandler import AragornHandler
from pdm_utils.classes.trnascansehandler import TRNAscanSEHandler
from pdm_utils.functions import basic

# Amino acids that we allow in the database
MYSQL_AMINO_ACIDS = {"Ala", "Arg", "Asn", "Asp", "Cys", "fMet", "Gln", "Glu",
                     "Gly", "His", "Ile", "Ile2", "Leu", "Lys", "Met", "Phe",
                     "Pro", "Pyl", "SeC", "Ser", "Stop", "Thr", "Trp", "Tyr",
                     "Val", "OTHER"}

# Amino acids that Genbank allows in the product field
GENBANK_AMINO_ACIDS = {"Ala", "Arg", "Asn", "Asp", "Cys", "Gln", "Glu", "Gly",
                       "His", "Ile", "Leu", "Lys", "Met", "Phe", "Pro", "Ser",
                       "Thr", "Trp", "Tyr", "Val", "OTHER"}

# Additional amino acids that SEA-PHAGES may annotate in the note field
SPECIAL_AMINO_ACIDS = {"fMet": "fMet", "Ile2": "Ile2", "Pyl": "Pyl",
                       "SeC": "SeC", "Sec": "SeC", "Term": "Stop",
                       "Stop": "Stop", "Sup": "Stop", "Undet": "OTHER"}

# Extracts amino acid from product field
PRODUCT_REGEX = re.compile("tRNA-(\w+)")

# Extracts amino acid and anticodon from note field for Aragorn-determinate
# or tRNAscan-SE- determinate or indeterminate tRNAs
NOTE_STANDARD_REGEX = re.compile("tRNA-(\w+)\s?\((\w+)\)")

# Extracts amino acid possibilities and anticodon from note field for
# Aragorn-indeterminate tRNAs
NOTE_SPECIAL_REGEX = re.compile("tRNA-\?\((\w+)\|(\w+)\)\s?\((\w+)\)")


class Trna:
    def __init__(self):
        """
        Constructor method for a tRNA object.
        """
        # The only attribute a tRNA really needs to have is a sequence
        self.seq = Seq("", IUPAC.ambiguous_dna)
        self.length = 0

        # Information about the tRNA with respect to its parent genome
        self.genome_id = ""     # Identifier for the parent genome
        self.genome_length = -1
        self.start = -1         # Start coord in parent genome, 0-indexed
        self.stop = -1          # Stop coord in parent genome, 0-indexed
        self.coordinate_format = ""
        self.orientation = ""   # "F" or "R", etc. - relative to parent genome

        # MySQL tRNAs will also need these:
        self.id = ""            # Identifier for this gene
        self.name = ""          # tRNA number in the parent genome
        self.amino_acid = "OTHER"   # Which aa. is this tRNA charged with?
        self.anticodon = "nnn"  # Which anticodon is at bottom of the C-loop?
        self.structure = ""     # What is the predicted secondary structure?

        # Genbank-formatted flat files should also have these attributes:
        self.seqfeature = None  # BioPython SeqFeature object for this tRNA
        self.locus_tag = ""     # Gene ID comprised of PhageID and Gene name
        self._locus_tag_num = ""
        self._gene_num = ""
        self.parts = 0          # How many regions does the tRNA have
        self.gene = ""          # Gene number parsed from feature
        self.product = ""       # Raw product field
        self.note = ""          # Raw note field

        # Aragorn data
        self.aragorn_data = None

        # tRNAscan-SE data
        self.trnascanse_data = None

        # Which program(s) support the annotation?
        self.sources = set()

        # Which program to check for valid anticodon?
        self.use = None

        # Useful for processing data from various sources:
        self.evaluations = list()
        self.type = ""
        self._start_stop_orient_id = tuple()
        self._end_orient_id = tuple()
        self._start_end_id = tuple()

    # TODO: create base feature class
    def set_locus_tag(self, tag="", delimiter="_", check_value=None):
        """
        Populate the locus_tag and parse the locus_tag number.
        :param tag: Input locus_tag data
        :type tag: str
        :param delimiter: Value used to split locus_tag data
        :type delimiter: str
        :param check_value: Genome name or other value that will be
        used to parse the locus_tag to identify the feature number
        :type check_value: str
        """
        # List of delimiters commonly found in locus tags. "_" is most common
        delims = ["_", "-", "."]

        # If tag is None, locus_tag and _locus_tag_num remain default ""
        if tag is None:
            return

        # Else, we'll set the tag directly, then decide how to parse number
        self.locus_tag = tag

        # If delimiter is None, choose based on what appears in locus_tag
        if delimiter is None:
            delimiter = basic.choose_most_common(tag, delims)

        # If no check_value was given, use self.genome_id
        if check_value is None:
            check_value = self.genome_id

        # Compile a search pattern and split the incoming tag on delimiter
        pattern = re.compile(check_value.lower())
        parts = tag.split(delimiter)

        # Attempt to parse locus tag number from the locus tag
        parsed_value = None
        for i in range(len(parts)):
            # Search each part of the locus tag for check_value
            search_result = pattern.search(parts[i].lower())
            if search_result is not None:
                if i == len(parts) - 1:
                    # If the check_value was found in last part of locus tag,
                    # value is probably everything after check_value
                    parsed_value = parts[i][search_result.end():]
                else:
                    # If the check_value was found anywhere else in locus tag,
                    # value is probably found in the next index
                    parsed_value = parts[-1]
                # No need to keep searching if we found a match
                break

        if parsed_value is None:
            # Last-ditch effort to find locus tag number
            parsed_value = parts[-1]

        # Remove generic 'gp' or 'orf' prefixes
        if parsed_value.lower().startswith("gp"):
            parsed_value = parsed_value[2:]
        elif parsed_value.lower().startswith("orf"):
            parsed_value = parsed_value[3:]
        else:
            pass

        # If the value we parsed is a digit, populate _locus_tag_num
        left, right = basic.split_string(parsed_value)
        if right != "":
            self._locus_tag_num = right
        else:
            self._locus_tag_num = parsed_value

    # TODO: create base feature class
    def set_name(self, value=None):
        """
        Set the feature name. Ideally, the name of the CDS will be an
        integer. This information can be stored in multiple fields in
        the GenBank-formatted flat file. The name is derived from one
        of several qualifiers.
        :param value: Indicates a value that should be used to
        directly set the name regardless of the 'gene' and
        '_locus_tag_num' attributes.
        :type value: str
        """
        # The CDS feature number may be stored in several places in the
        # gene, product, note, function, and locus_tag qualifiers of a
        # flat file, but it is unpredictable. Not all of these qualifiers
        # are always present.
        # 1. PECAAN Draft and new SEA-PHAGES Final:
        #    The 'gene' qualifier should be present and contain an integer.
        # 2. SEA-PHAGES Final in GenBank:
        #    The 'gene' qualifier may or may not be present, and
        #    it may or may not have an integer.
        #    The 'locus_tag' qualifier may or may not be present,
        #    and may or may not have an integer.
        # 3. Non-SEA-PHAGES in GenBank:
        #    The 'gene' qualifier may or may not be present, and
        #    it may or may not have an integer.
        #    The 'locus_tag' qualifier may or may not be present,
        #    and may or may not have an integer.
        if value is None:
            name = ""
            list1 = ["_locus_tag_num", "_gene_num"]

            # First see if any num attributes have a float.
            x = 0
            while name == "" and x < len(list1):
                value = getattr(self, list1[x])
                if basic.is_float(value):
                    name = value
                x += 1

            # Second see if any num attributes have non-empty values.
            # At this point, it's very unpredictable. Values could be like
            # 'terL' in the gene or like '10a' in the locus_tag.
            if name == "":
                list2 = ["gene"]
                list2.extend(list1)
                y = 0
                while name == "" and y < len(list2):
                    value = getattr(self, list2[y])
                    if value != "":
                        name = value
                    y += 1
        self.name = value

    # TODO: create base feature class
    def set_gene(self, value, delimiter=None, prefix_set=None):
        """
        Set the gene attribute.
        :param value: Gene data to parse. Also passed to set_num().
        :type value: str
        :param delimiter: Passed to set_num().
        :type delimiter: str
        :param prefix_set: Passed to set_num().
        :type prefix_set: set
        """
        value = value.strip()
        self.gene = value
        self._gene_num = value
        self.set_num("_gene_num", value,
                     delimiter=delimiter, prefix_set=prefix_set)

    # TODO: create base feature class
    def set_num(self, attr, description, delimiter=None, prefix_set=None):
        """
        Set a number attribute from a description.
        :param attr: Attribute to set the number.
        :type attr: str
        :param description: Description data from which to parse the number.
        :type description: str
        :param delimiter: Value used to split the description data.
        :type delimiter: str
        :param prefix_set: Valid possible delimiters in the description.
        :type prefix_set: set
        """
        # Used to set product_num, function_num, note_num, gene_num
        # Sometimes feature number is stored within the description
        # (e.g. 'gp10; terminase' or 'terminase; gp10')
        # List of delimiters found in description fields
        # ";" is probably the most common.
        delimiters = [";", ","]
        if delimiter is None:
            delimiter = basic.choose_most_common(description, delimiters)
        if prefix_set is None:
            prefix_set = {"gp", "orf", ""}

        # Iterate through the list of strings, and select the first
        # string that looks like a gene number.
        split_value = description.split(delimiter)
        num = ""
        x = 0
        while (num == "" and x < len(split_value)):
            string = split_value[x]
            string = string.strip()
            left, right = basic.split_string(string)
            # Possible returns:
            # 1. left is alpha and right is number or float (gp10 = gp, 10),
            # 2. string is alpha (terminase = terminase, ""),
            # 3. string is number (10 = "", 10).
            if left.lower() in prefix_set:
                num = right
            x += 1
        setattr(self, attr, num)

    # TODO: create base feature class - use this version of the method unless
    #  Travis objects, because it's more Pythonic.
    def get_begin_end(self):
        """
        Accesses feature coordinates in transcription begin-end format.
        :return: (begin, end)
        """
        # Get a copy of the orientation in fr_long format:
        orientation = basic.reformat_strand(self.orientation, "fr_long")

        if orientation == "forward":
            # Rightward transcribed gene
            begin, end = self.start, self.stop
        elif orientation == "reverse":
            # Leftward transcribed gene
            begin, end = self.stop, self.start
        else:
            # Unexpected orientation
            begin = end = -1

        return begin, end

    # TODO: create base feature class - use this version of the method unless
    #  Travis objects, because it's documented more cleanly and avoids using
    #  keyword 'format'.
    def set_orientation(self, value, fmt, case=False):
        """
        Set the orientation based on the indicated format.
        :param value: orientation value
        :type value: int or str
        :param fmt: how orientation should be formatted
        :type fmt: str
        :param case: whether to capitalize the first letter of
        orientation
        :type case: bool
        :return:
        """
        self.orientation = basic.reformat_strand(value, fmt, case)

    # TODO: create base feature class
    def set_location_id(self):
        """
        Create identifier tuples containing feature location data. For
        this method we only care about gene boundaries and will ignore
        any multi-part elements to the gene.
        :return:
        """
        # Tuple with ALL relevant information
        self._start_stop_orient_id = (self.start, self.stop, self.orientation)

        begin, end = self.get_begin_end()

        # If only end coordinate is used, we need to know what orientation the
        # gene is in - two genes could theoretically be transcribed in opposed
        # directions and have the same stop position
        self._end_orient_id = (end, self.orientation)

        # Or we can have transcription begin/end coordinates, in which case
        # the direction can be gleaned from these values
        self._start_end_id = (begin, end)

    # TODO: create base feature class - use this version of the method unless
    #  Travis objects, because it has slightly improved logic.
    def reformat_start_and_stop(self, fmt):
        """
        Convert existing start and stop coordinates to the indicated
        new format; also updates the coordinate format attribute to
        reflect any change.
        :param fmt: the new desired coordinate format
        :type fmt: str
        :return:
        """
        # Only need to do work if the new coordinate format differs from
        # the existing coordinate format
        if fmt != self.coordinate_format:
            self.start, self.stop = basic.reformat_coordinates(
                self.start, self.stop, self.coordinate_format, fmt)
            self.coordinate_format = fmt

    # TODO: create base feature class - fully equivalent to version in Cds,
    #  but documented differently.
    def set_nucleotide_length(self, use_seq=False):
        """
        Set the nucleotide length of this gene feature.
        :param use_seq: whether to use the Seq feature to calculate
        nucleotide length of this feature
        :type use_seq: bool
        :return:
        """
        if use_seq:
            self.length = len(self.seq)
        else:
            if self.coordinate_format == "0_half_open":
                # Python coordinate format
                self.length = self.stop - self.start
            elif self.coordinate_format == "1_closed":
                # Genbank coordinate format
                self.length = self.stop - self.start + 1
            else:
                # Unknown coordinate format
                self.length = -1

    # TODO: create base feature class - use this version of the method unless
    #  Travis objects, because the logic is slightly better and it raises an
    #  Error if something is wrong with given seq.
    def set_nucleotide_sequence(self, value=None, parent_genome_seq=None):
        """
        Set this feature's nucleotide sequence
        :param value: sequence
        :type value: str or Seq
        :param parent_genome_seq: parent genome sequence
        :type parent_genome_seq: Seq
        :raise: ValueError
        :return:
        """
        # If seq is given we'll try to use that
        if value is not None:
            # If seq is a BioPython Seq object, use it directly
            if isinstance(value, Seq):
                self.seq = value.upper()
            # If seq is a string, we can coerce it to a BioPython Seq object
            elif isinstance(value, str):
                self.seq = Seq(value.upper(), IUPAC.ambiguous_dna)
            # If seq is something else entirely, we cannot use it
            else:
                pass
                # raise ValueError(f"tRNA.seq cannot use type '{type(value)}'")
        # If instead a parent_genome_seq is given, we'll try to use that
        elif parent_genome_seq is not None and self.seqfeature is not None:
            try:
                self.seq = self.seqfeature.extract(parent_genome_seq)
            # TODO: tighten exception clause, OR let the error pass to a higher
            #  level in the call stack
            except:
                # Leave as default
                pass
        # If neither is given, leave self.seq as default
        else:
            pass

    # TODO: create base feature class - fully equivalent to the version in Cds,
    #  but better documented.
    def set_seqfeature(self, type=None):
        """
        Create a SeqFeature object with which to populate the
        `seqfeature` attribute.
        :return:
        """
        if type is None:
            type = self.type

        # SeqFeature coordinates are 0-based half-open
        start, stop = basic.reformat_coordinates(
            self.start, self.stop, self.coordinate_format, "0_half_open")

        # SeqFeature orientation is (-1, 1) instead of ("R", "F")
        strand = basic.reformat_strand(self.orientation, "numeric")

        self.seqfeature = self.create_seqfeature(type, start, stop, strand)

    def create_seqfeature(self, type, start, stop, strand):
        # Standard genes will have start < stop
        if start <= stop:
            seqfeature = SeqFeature(FeatureLocation(start, stop),
                                         strand=strand, type=type)
        # Wrap-around genes will have stop < start
        else:
            seqfeature = SeqFeature(CompoundLocation(
                [FeatureLocation(start, self.genome_length),
                 FeatureLocation(0, stop)]), strand=strand, type=type)
        # Add feature qualifiers
        seqfeature.qualifiers = self.get_qualifiers(type)

        return seqfeature

    def get_qualifiers(self, type):
        """
        Helper function that uses tRNA data to populate the qualifiers
        attribute of `seqfeature`.
        :return: qualifiers OrderedDict()
        """
        qualifiers = OrderedDict()
        if type == "tRNA":
            qualifiers["gene"] = [self.name]
            if self.locus_tag != "":
                qualifiers["locus_tag"] = [self.locus_tag]

            if self.amino_acid not in GENBANK_AMINO_ACIDS:
                amino_acid = "OTHER"
            else:
                amino_acid = self.amino_acid

            qualifiers["product"] = [f"tRNA-{amino_acid}"]
            qualifiers["note"] = [f"tRNA-{amino_acid} ({self.anticodon})"]
        if type == "gene":
            qualifiers["gene"] = [self.name]
            qualifiers["locus_tag"] = [self.locus_tag]

        return qualifiers

    def run_aragorn(self):
        """
        Uses an AragornHandler object to negotiate the flow of
        information between this object and Aragorn.
        :return:
        """
        if self.id is not None and self.id != "":
            identifier = self.id
        else:
            identifier = "aragorn"

        if len(self.seq) > 0:
            ah = AragornHandler(identifier, str(self.seq))
            ah.write_fasta()
            ah.run_aragorn()            # search (linear) self.seq for tRNAs
            ah.read_output()
            ah.parse_trnas()
            if ah.trna_tally == 1:
                self.aragorn_data = ah.trnas[0]
                self.sources.add("aragorn")
            # else:
                # print(f"Aragorn found {ah.trna_tally} tRNAs in this region.")
        else:
            print("Cannot run Aragorn on 0-length sequence.")

    def run_trnascanse(self):
        """
        Uses a TRNAscanSEHandler object to negotiate the flow of
        information between this object and tRNAscan-SE.
        :return:
        """
        if self.id is not None and self.id != "":
            identifier = self.id
        else:
            identifier = "trnascanse"

        if len(self.seq) > 0:
            th = TRNAscanSEHandler(identifier, str(self.seq))
            th.write_fasta()
            th.run_trnascanse()
            th.read_output()
            th.parse_trnas()
            if th.trna_tally == 1:
                self.trnascanse_data = th.trnas[0]
                self.sources.add("trnascan")
            # else:
                # print(f"tRNAscan-SE found {th.trna_tally} tRNAs in this region.")
        else:
            print("Cannot run tRNAscan-SE on 0-length sequence.")

    def set_amino_acid(self, value):
        """
        Sets the `amino_acid` attribute using the indicated value.
        :param value: the Amino acid to be used
        :type value: str
        :raise: ValueError
        :return:
        """
        self.amino_acid = value

    def parse_amino_acid(self):
        """
        Attempts to parse the `amino_acid` attribute from the `product`
        and `note` attributes.
        :return:
        """
        results = PRODUCT_REGEX.findall(self.product)
        # If we got a single item (e.g. the amino acid)
        if len(results) == 1:
            # Begin using it as the amino acid
            amino_acid = results[0]
            # If amino acid we got is "OTHER", we need to check the note
            # field for a more specific amino acid
            if amino_acid == "OTHER":
                results = NOTE_STANDARD_REGEX.findall(self.note)
                # If we got two items (e.g. the amino acid and anticodon)
                # and the amino acid is in our allowed special amino acids
                if len(results) != 0 and results[0][0] in SPECIAL_AMINO_ACIDS:
                    # Use the re-formatted special amino acid
                    amino_acid = SPECIAL_AMINO_ACIDS[results[0][0]]
                # Else, we will leave amino acid as "OTHER"
            # Else, there is nothing else to do but set the attribute
            self.amino_acid = amino_acid
        # Else, the product is not formatted properly and will fail checks

    def set_anticodon(self, value):
        """
        Sets the `anticodon` attribute using the indicated value.
        :param value: the anticodon to use for this tRNA
        :type value: str
        :return:
        """
        self.anticodon = value.lower()

    def parse_anticodon(self):
        """
        Attempts to parse the `anticodon` attribute from the `note`
        attribute.
        :return:
        """
        results = NOTE_STANDARD_REGEX.findall(self.note)
        # If we got two items (e.g. the amino acid and anticodon)
        if len(results) == 1:
            self.anticodon = results[0][-1].lower().replace('u', 't')
        else:
            results = NOTE_SPECIAL_REGEX.findall(self.note)
            # If we got three items (e.g. the two amino acid choices and
            # part of anticodon)
            if len(results) == 1:
                self.anticodon = results[0][-1].lower().replace('u', 't')
        # If the regex fails to parse an anticodon using either regex, it
        # will be left as "nnn"

    def set_structure(self, value):
        """
        Set the secondary structure string so downstream users can
        easily display the predicted fold of this tRNA.
        :param value: the string to use as the secondary structure
        :type value: str
        :return:
        """
        self.structure = value

    # Evaluations
    # TODO: create base feature class - fully equivalent to version in Cds,
    #  but documented slightly differently.
    def set_eval(self, eval_id, definition, result, status):
        """
        Constructs and adds and Evaluation object to this feature's list of
        evaluations.
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param definition: description of the evaluation
        :type definition: str
        :param result: description of the evaluation outcome
        :type result: str
        :param status: overall outcome of the evaluation
        :type status: str
        :return:
        """
        evl = evaluation.Evaluation(eval_id, definition, result, status)
        self.evaluations.append(evl)

    # TODO: create base feature class - nearly identical to Cds version,
    #  but exception clause is appropriately narrow.
    def check_attribute(self, attribute, check_set, expect=False, eval_id=None,
                        success="correct", fail="error", eval_def=None):
        """
        Checks whether the indicated feature attribute is present in
        the given check_set. Uses expect to determine whether the
        presence (or lack thereof) is an error, or correct.
        :param attribute: the gene feature attribute to evaluate
        :type attribute: str
        :param check_set: set of reverence values
        :type check_set: set
        :param expect: whether the attribute's value is expected to be
        in the reference set
        :type expect: bool
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        try:
            value = getattr(self, attribute)
        except AttributeError:
            value = None

        if value is not None:
            value_short = basic.truncate_value(str(value), 30, "...")
            result = f"The '{attribute}' value '{value_short}' is "

            # Returns a boolean
            outcome = basic.check_value_expected_in_set(
                value, check_set, expect)

            if outcome:
                result += "valid."
                status = success
            else:
                result += "not valid."
                status = fail
        else:
            result = f"'{attribute}' is not a valid attribute to be evaluated."
            status = "untested"

        definition = f"Check the value of the '{attribute}' attribute for " \
                     f"{self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    # TODO: create base feature class - Travis' version assumes value can't
    #  be None; if wrong, that method will fail because None can't be <=> x.
    def check_magnitude(self, attribute, expect, ref_value, eval_id=None,
                        success="correct", fail="error", eval_def=None):
        """
        Check that the magnitude of a numerical attribute meets
        expectations.
        :param attribute: the gene feature attribute to evaluate
        :type attribute: str
        :param expect: symbol designating direction of magnitude (>=<)
        :type expect: str
        :param ref_value: numerical value for comparison
        :type ref_value: int, float, datetime
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        try:
            value = getattr(self, attribute)
        except AttributeError:
            value = None

        if value is not None:
            result = f"The '{attribute}' value '{value}' is "
            if value > ref_value:
                compare = ">"
                result += "greater than "
            elif value == ref_value:
                compare = "="
                result += "equal to "
            else:
                compare = "<"
                result += "less than "
            result += f"{ref_value}, which is "
            if compare == expect:
                result += "expected."
                status = success
            else:
                result += "not expected."
                status = fail
        else:
            result = f"'{attribute}' is not a valid attribute to be evaluated."
            status = "untested"

        definition = f"Check the magnitude of the '{attribute}' attribute " \
                     f"for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    def check_sources(self, eval_id=None, success="correct", fail="error",
                      eval_def=None):
        """
        Check that this tRNA's DNA sequence can successfully turn up
        a tRNA when run through Aragorn and tRNAscan-SE.
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        self.run_aragorn()
        self.run_trnascanse()

        result = f"This tRNA gene's DNA sequence "

        if self.sources == set():
            # Neither program found a tRNA in this region
            self.use = None
            result += "does not appear to correspond to a tRNA."
            status = fail
        elif self.sources == {"aragorn"}:
            # Aragorn found a tRNA in this region but tRNAscan-SE did not
            self.use = "aragorn"
            result += "corresponds to a tRNA in Aragorn."
            status = success
        elif self.sources == {"trnascan"}:
            # tRNAscan-SE found a tRNA in this region but Aragorn did not
            self.use = "trnascan"
            result += "corresponds to a tRNA in tRNAscan-SE."
            status = success
        elif self.sources == {"aragorn", "trnascan"}:
            # Both programs found a tRNA in this region
            self.use = "both"
            result += "corresponds to a tRNA in both Aragorn and tRNAscan-SE."
            status = success
        else:
            # Something unexpected happened...
            self.use = None
            result += "gave rise to something unexpected."
            status = fail
            print(f"Invalid tRNA validation source in ({self.sources}).")

        definition = f"Check that there is bioinformatic support for the " \
                     f"existence of {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    # TODO: create base feature class - this version avoids overloading
    #  built-in format()
    def check_orientation(self, fmt="fr_short", case=True, eval_id=None,
                          success="correct", fail="error", eval_def=None):
        """
        Check that the orientation is set appropriately.
        :param fmt: indicates how coordinates should be formatted
        :type fmt: str
        :param case: indicates whether orientation data should be cased
        :type case: bool
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        expect = basic.reformat_strand(self.orientation, fmt, case)
        result = f"The orientation is {self.orientation}, and it is formatted "

        if self.orientation == expect:
            result += "correctly."
            status = success
        else:
            result += "incorrectly."
            status = fail

        definition = f"Check if the orientation is set correctly for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    def check_orientation_correct(self, fmt="fr_short", case=True,
                                  eval_id=None, success="correct",
                                  fail="error", eval_def=None):
        """
        Check that the orientation agrees with the Aragorn and/or
        tRNAscan-SE predicted orientation. If Aragorn/tRNAscan-SE
        report a forward orientation, it means they agree with the
        annotated orientation. If they report reverse orientation,
        they think the annotation is backwards.
        :param fmt: indicates how coordinates should be formatted
        :type fmt: str
        :param case: indicates whether orientation data should be cased
        :type case: bool
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        orientation = basic.reformat_strand(self.orientation, fmt, case)
        result = f"The annotated orientation '{orientation}' "

        if self.use == "both":
            a_orient = basic.reformat_strand(
                self.aragorn_data["Orientation"], fmt, case)
            t_orient = basic.reformat_strand(
                self.trnascanse_data["Orientation"], fmt, case)
            if a_orient == t_orient == "F":
                result += "matches Aragorn and tRNAscan-SE predictions."
                status = success
            elif a_orient == t_orient == "R":
                result += "is backwards relative to Aragorn and tRNAscan-SE " \
                          "predictions."
                status = fail
            else:
                result += "cannot be sensibly checked, as Aragorn and " \
                          "tRNAscan-SE disagree."
                status = "unchecked"
        elif self.use == "aragorn":
            a_orient = basic.reformat_strand(
                self.aragorn_data["Orientation"], fmt, case)
            if a_orient == "F":
                result += "matches the Aragorn prediction."
                status = success
            elif a_orient == "R":
                result += "is backwards relative to Aragorn prediction."
                status = fail
            else:
                result += "cannot be sensibly checked, as Aragorn did" \
                          "something unexpected"
                status = "unchecked"
        elif self.use == "trnascan":
            t_orient = basic.reformat_strand(
                self.trnascanse_data["Orientation"], fmt, case)
            if t_orient == "F":
                result += "matches the tRNAscan-SE prediction."
                status = success
            elif t_orient == "R":
                result += "is backwards relative to tRNAscan-SE prediction."
                status = fail
            else:
                result += "cannot be sensibly checked, as tRNAscan-SE did" \
                          "something unexpected"
                status = "unchecked"
        # If neither predict a tRNA, we presume the annotation is wrong.
        else:
            result += "is at odds with expectations (no tRNA)."
            status = fail

        definition = f"Check whether the annotated orientation for {self.locus_tag} ({self.id})" \
                     f" matches the Aragorn/tRNAscan-SE prediction(s)."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    def check_amino_acid_valid(self, eval_id=None, success="correct",
                               fail="error", eval_def=None):
        """
        Checks that the amino acid that has been annotated for this
        tRNA is in the set of amino acids that we have opted to allow
        in the MySQL database.
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        result = f"The annotated amino acid '{self.amino_acid}' is "

        if self.amino_acid in MYSQL_AMINO_ACIDS:
            result += "valid."
            status = success
        else:
            result += "invalid."
            status = fail

        definition = f"Check that the annotated amino acid is in the allowed" \
                     f" set for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    def check_amino_acid_correct(self, eval_id=None, success="correct",
                                 fail="error", eval_def=None):
        """
        Checks that the amino acid that has been annotated for this
        tRNA agrees with the Aragorn and/or tRNAscan-SE prediction(s).
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        result = f"The annotated isotype ({self.amino_acid}) "

        # If both programs found a tRNA here
        if self.use == "both":
            # All Aragorn isotypes are valid in our database except undet
            a_isotype = self.aragorn_data["AminoAcid"]
            if "|" in a_isotype:
                a_isotype = "OTHER"
            # Some tRNAscan-SE isotypes need to be converted to allowed value
            t_isotype = self.trnascanse_data["AminoAcid"]
            if t_isotype not in MYSQL_AMINO_ACIDS:
                if t_isotype in SPECIAL_AMINO_ACIDS:
                    t_isotype = SPECIAL_AMINO_ACIDS[t_isotype]
                else:
                    t_isotype = "OTHER"
            # if Aragorn and tRNAscan-SE are the same
            if a_isotype == t_isotype == self.amino_acid:
                result += f"is consistent with Aragorn ({a_isotype}) and " \
                          f"tRNAscan-SE ({t_isotype})."
                status = success
            # if Aragorn is Met and tRNAscan-se is consistent
            elif a_isotype == "Met" and t_isotype in ("Met", "fMet", "Ile2"):
                if self.amino_acid in (a_isotype, t_isotype):
                    result += f"is consistent with Aragorn ({a_isotype}) " \
                              f"and tRNAscan-SE ({t_isotype})."
                    status = success
                else:
                    result += f"is inconsistent with Aragorn ({a_isotype}) " \
                              f"and tRNAscan-SE ({t_isotype})."
                    status = fail
            # if tRNAscan-SE is Sup and Aragorn is consistent
            elif a_isotype in ("Pyl", "Stop") and t_isotype == "Sup":
                if self.amino_acid in (a_isotype, t_isotype):
                    result += f"is consistent with Aragorn ({a_isotype}) " \
                              f"and tRNAscan-SE ({t_isotype})."
                    status = success
                else:
                    result += f"is inconsistent with Aragorn ({a_isotype}) " \
                              f"and tRNAscan-SE ({t_isotype})."
                    status = fail
            else:
                result += f"cannot be sensibly checked because Aragorn " \
                          f"({a_isotype}) prediction differs from " \
                          f"tRNAscan-SE ({t_isotype})."
                status = "unchecked"
        # If only Aragorn found a tRNA here
        elif self.use == "aragorn":
            a_isotype = self.aragorn_data["AminoAcid"]
            if "|" in a_isotype:
                a_isotype = "OTHER"
            t_isotype = "no tRNA"
            if a_isotype == self.amino_acid:
                result += f"is consistent with Aragorn ({a_isotype}) and " \
                          f"tRNAscan-SE ({t_isotype})."
                status = success
            else:
                result += f"is inconsistent with Aragorn ({a_isotype}) " \
                          f"and tRNAscan-SE ({t_isotype})."
                status = fail
        # If only tRNAscan-SE found a tRNA here
        elif self.use == "trnascan":
            a_isotype = "no tRNA"
            t_isotype = self.trnascanse_data["AminoAcid"]
            if t_isotype not in MYSQL_AMINO_ACIDS:
                if t_isotype in SPECIAL_AMINO_ACIDS:
                    t_isotype = SPECIAL_AMINO_ACIDS[t_isotype]
                else:
                    t_isotype = "OTHER"
            if t_isotype == self.amino_acid:
                result += f"is consistent with Aragorn ({a_isotype}) and " \
                          f"tRNAscan-SE ({t_isotype})."
                status = success
            else:
                result += f"is inconsistent with Aragorn ({a_isotype}) " \
                          f"and tRNAscan-SE ({t_isotype})."
                status = fail
        # If neither program found a tRNA here
        else:
            result += "is inconsistent with Aragorn (no tRNA) and " \
                      "tRNAscan-SE (no tRNA)."
            status = fail

        definition = f"Check that the annotated amino acid is consistent " \
                     f"with the prediction(s) made by Aragorn and/or " \
                     f"tRNAscan-SE for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    def check_anticodon_valid(self, eval_id=None, success="correct",
                              fail="error", eval_def=None):
        """
        Checks that the anticodon conforms to the expected length
        (2-4) and alphabet ("a", "c", "g", "t") or is "nnn".
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        result = f"The anticodon '{self.anticodon}' is "
        expect = ["a", "c", "g", "t"]
        valid = [letter in expect for letter in self.anticodon]
        if len(self.anticodon) == 3:
            if sum(valid) == 3:
                result += "valid."
                status = success
            elif self.anticodon == "nnn":
                result += "valid."
                status = success
            else:
                result += "invalid."
                status = fail
        elif len(self.anticodon) in [2, 4]:
            if sum(valid) == len(self.anticodon):
                result += "valid."
                status = success
            else:
                result += "invalid."
                status = fail
        else:
            result += "invalid."
            status = fail

        definition = f"Check that the anticodon is valid for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    def check_anticodon_correct(self, eval_id=None, success="correct",
                              fail="error", eval_def=None):
        """
        Checks that the annotated anticodon agrees with the prediction
        by Aragorn or tRNAscan-SE.
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        result = f"The anticodon '{self.anticodon}' "

        # If both programs found a tRNA here
        if self.use == "both":
            a_anti = self.aragorn_data["Anticodon"]
            t_anti = self.trnascanse_data["Anticodon"]
            if a_anti == t_anti == self.anticodon:
                result += f"is consistent with Aragorn ({a_anti}) and " \
                          f"tRNAscan-SE ({t_anti})."
                status = success
            elif a_anti == t_anti != self.anticodon:
                result += f"is inconsistent with Aragorn ({a_anti}) and " \
                          f"tRNAscan-SE ({t_anti})."
                status = fail
            else:
                result += f"cannot be sensibly checked because Aragorn " \
                          f"({a_anti}) prediction differs from tRNAscan-SE " \
                          f"({t_anti})."
                status = "unchecked"
        # If only Aragorn found a tRNA here
        elif self.use == "aragorn":
            a_anti = self.aragorn_data["Anticodon"]
            t_anti = "no tRNA"
            if a_anti == self.anticodon:
                result += f"is consistent with Aragorn ({a_anti}) and " \
                          f"tRNAscan-SE ({t_anti})."
                status = success
            else:
                result += f"is inconsistent with Aragorn ({a_anti}) and " \
                          f"tRNAscan-SE ({t_anti})."
                status = fail
        # If only tRNAscan-SE found a tRNA here
        elif self.use == "trnascan":
            a_anti = "no tRNA"
            t_anti = self.trnascanse_data["Anticodon"]
            if t_anti == self.anticodon:
                result += f"is consistent with Aragorn ({a_anti}) and " \
                          f"tRNAscan-SE ({t_anti})."
                status = success
            else:
                result += f"is inconsistent with Aragorn ({a_anti}) and " \
                          f"tRNAscan-SE ({t_anti})."
                status = fail
        # If neither program found a tRNA here
        else:
            result += "is inconsistent with Aragorn (no tRNA) and " \
                      "tRNAscan-SE (no tRNA)."
            status = fail

        definition = f"Check that the annotated anticodon agrees with the " \
                     f"Aragorn or tRNAscan-SE anticodon for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    # TODO: create base feature class
    def check_locus_tag_structure(self, check_value=None, only_typo=False,
                                  prefix_set=set(), case=True, eval_id=None,
                                  success="correct", fail="error", eval_def=None):
        """Check if the locus_tag is structured correctly.

        :param check_value:
            Indicates the genome id that is expected to be present.
            If None, the 'genome_id' parameter is used.
        :type check_value: str
        :param only_typo:
            Indicates if only the genome id spelling should be evaluated.
        :type only_typo: bool
        :param prefix_set:
            Indicates valid common prefixes, if a prefix is expected.
        :type prefix_set: set
        :param case:
            Indicates whether the locus_tag is expected to be capitalized.
        :type case: bool
        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """
        if check_value is None:
            check_value = self.genome_id

        results = []
        if only_typo:
            pattern = re.compile(check_value.lower())
            search_result = pattern.search(self.locus_tag.lower())
            if search_result == None:
                results.append("The genome ID is missing.")
        else:
            # Expected structure: SEA_TRIXIE_20
            parts = self.locus_tag.split("_")
            if case:
                if self.locus_tag != self.locus_tag.upper():
                    results.append("The capitalization is incorrect.")
            if len(parts) == 3:
                if prefix_set is not None:
                    if parts[0].upper() not in prefix_set:
                        results.append("The prefix is missing.")
                if parts[1].upper() != check_value.upper():
                    results.append("The genome ID is missing.")
                if not parts[2].isdigit():
                    results.append("The feature number is missing.")
            else:
                results.append("The locus_tag does not contain three parts.")
        result = f"The locus_tag qualifier is {self.locus_tag}. It is "
        if len(results) == 0:
            result = result + "structured correctly."
            status = success
        else:
            result = result + "not structured correctly. " + " ".join(results)
            status = fail
        definition = f"Check if the locus_tag qualifier is structured correctly for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    # TODO: create base feature class - fully equivalent to the version in Cds,
    #  but exception clause is appropriately narrow.
    def check_gene_structure(self, eval_id=None, success="correct",
                             fail="error", eval_def=None):
        """
        Check that the gene qualifier contains an integer.
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        try:
            value = int(self.gene)
        except ValueError:
            value = self.gene

        result = f"The gene qualifier is {self.gene}. It"
        if isinstance(value, int):
            result += "contains an integer, as expected."
            status = success
        else:
            result += "does ont contain an integer, which is not expected."
            status = fail

        definition = f"Check if the gene qualifier contains an integer for " \
                     f"{self.locus_tag}({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    # TODO: create base feature class
    def check_compatible_gene_and_locus_tag(self, eval_id=None,
                                            success="correct", fail="error",
                                            eval_def=None):
        """
        Check that gene and locus_tag attributes contain identical
        numbers.
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        result = f"The numbers in the gene ({self.gene}) and " \
                 f"locus_tag ({self.locus_tag}) qualifiers are "
        if self.gene == self._locus_tag_num:
            result += "consistent."
            status = success
        else:
            result += "not consistent."
            status = fail

        definition = f"Check if the gene and locus_tag numbers are " \
                     f"consistent for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    def check_length(self, eval_id=None, success="correct", fail="error",
                     eval_def=None):
        """
        Checks that the tRNA is in the expected range of lengths. The
        average tRNA gene is 70-90bp in length, but it is not uncommon
        to identify well-scoring tRNAs in the 60-100bp range.
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        result = f"The tRNA length ({self.length}) is "
        if self.length in range(60, 101):
            result += "in the expected range."
            status = success
        elif self.length < 60:
            result += "shorter than expected."
            status = fail
        else:
            result += "longer than expected."
            status = fail

        definition = f"Check if the tRNA length is in the expected range " \
                     f"for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    def check_coordinates(self, eval_id=None, success="correct", fail="error",
                          eval_def=None):
        """

        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        result = f"The tRNA coordinates ({self.start}, {self.stop}) "

        # We'll use Aragorn by default when available
        if self.use == "both" or self.use == "aragorn":
            start_offset = self.aragorn_data["Start"]
            aragorn_stop = self.aragorn_data["Stop"]
            stop_offset = self.length - aragorn_stop
            structure = self.aragorn_data["Structure"]
        # Use tRNAscan-SE if that's all we've got
        elif self.use == "trnascan":
            start_offset = self.trnascanse_data["Start"]
            trnascanse_stop = self.trnascanse_data["Stop"]
            stop_offset = self.length - trnascanse_stop
            structure = self.trnascanse_data["Structure"]
        else:
            start_offset = -99999     # arbitrary number that will never happen
            stop_offset = -99999      # arbitrary number that will never happen
            structure = ""

        # If tRNA is not forward-oriented, the start/stop offsets are reversed
        # because we put the real tRNA DNA sequence as the input, NOT the top
        # strand DNA sequence
        orientation = basic.reformat_strand(
            self.orientation, format="fr_short", case=True)
        if orientation == "R":
            start_offset, stop_offset = stop_offset, start_offset

        # Now check the offsets and adjust the structure as appropriate
        if start_offset == stop_offset == 0:
            result += "appear to be correct."
            status = success
        elif start_offset == stop_offset == -99999:
            result += f"should be investigated - neither Aragorn nor " \
                      f"tRNAscan-SE identify a tRNA in this position."
            status = fail
        else:
            result += "appear to be incorrect. Left coordinate "
            status = fail

            if start_offset == 0:
                result += "looks correct. "
            # Else we need to modify the structure to accommodate extra or
            # missing bases
            elif start_offset < 0:
                # tRNA is annotated shorter than Aragorn says it should be
                structure = structure[abs(start_offset):]
                result += f"should be moved left by {abs(start_offset)} " \
                          f"base(s). "
            else:
                # tRNA is annotated longer than Aragorn thinks it should be
                structure = ("." * start_offset) + structure
                result += f"should be moved right by {start_offset} base(s). "

            result += "Right coordinate "
            if stop_offset == 0:
                result += "looks correct."
            elif stop_offset < 0:
                # tRNA is annotated shorter than Aragorn says it should be
                structure = structure[:self.length]
                result += f"should be moved right by {abs(stop_offset)} " \
                          f"base(s)."
            else:
                # tRNA is annotated longer than Aragorn thinks it should be
                structure = structure + ("." * stop_offset)
                result += f"should be moved left by {stop_offset} base(s)."

        self.structure = structure

        definition = f"Check if the tRNA coordinates appear to match the " \
                     f"Aragorn or tRNAscan-SE prediction(s) for " \
                     f"{self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    def check_terminal_nucleotides(self, eval_id=None, success="correct",
                                   fail="warning", eval_def=None):
        """
        Checks that the tRNA ends with "CCA" or "CC" or "C".
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        if len(self.seq) > 4:
            result = f"The tRNA ends with '...{self.seq[-4:]}'. "
            # tRNAs must end in "NCCA" to function - some are this way naturally
            # The others must end in "NCC" or "NC" or N
            if self.seq.endswith("CCA"):
                result += "This is expected."
                status = success
            elif self.seq.endswith("CC"):
                result += "This is expected."
                status = success
            elif self.seq.endswith("C"):
                result += "This is expected."
                status = success
            else:
                result += " This is not expected."
                status = fail
        else:
            result = f"Cannot check terminal nucleotides on a sequence of " \
                     f"length {len(self.seq)}."
            status = "unchecked"

        definition = f"Check that the correct terminal nucleotide(s) are " \
                     f"present for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    def check_product_structure(self, eval_id=None, success="correct",
                                fail="error", eval_def=None):
        """
        Checks that the product field is formatted properly, and that
        the annotated amino acid is valid.
        Genbank enforces that all tRNA annotations that have a product
        field have it annotated as either 'tRNA-Xxx', where Xxx is one
        of the 20 standard amino acids, or 'tRNA-OTHER' for those tRNAs
        which decode a non-standard amino acid (e.g. SeC, Pyl, fMet).
        SEA-PHAGES may also append the anticodon parenthetically for a
        product field such as 'tRNA-Xxx(nnn)'.
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        result = f"The tRNA product is "
        if self.product != "":
            result += f"present ('{self.product}'). "

            # This regex will ignore the (nnn) anticodon if it's there; it
            # parses the amino acid prediction and makes sure it's valid.
            results = PRODUCT_REGEX.findall(self.product)

            # If the product field is formatted correctly, there should only
            # be a single result from the regex search
            if results is not None and len(results) != 1:
                result += f"Product is formatted properly. "
                status = success

            # If there is more or less than a single result, the product field
            # is not properly formatted
            else:
                result += "Product is not formatted properly."
                status = fail
        # If there is no product field, it's not formatted properly.
        else:
            result += "missing (''). "
            result += "Product is not formatted properly."
            status = fail

        definition = f"Check that the product is formatted properly for " \
                     f"{self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    def check_note_structure(self, eval_id=None, success="correct",
                             fail="error", eval_def=None):
        """
        Checks that the note field is formatted properly.

        Genbank does not enforce any standard for the note field.
        This means that a note does not have to exist.

        SEA-PHAGES note fields should look like 'tRNA-Xxx(nnn)'.
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        result = f"The tRNA note is "
        if self.note != "":
            result += f"present ('{self.note}'). "

            results1 = NOTE_STANDARD_REGEX.findall(self.note)
            results2 = NOTE_SPECIAL_REGEX.findall(self.note)

            if results1 is not None and len(results1) == 2:
                result += f"Note is formatted properly."
                status = success
            elif results2 is not None and len(results2) == 3:
                result += f"Note is formatted properly."
                status = success
            else:
                result += f"Note is not formatted properly."
                status = fail
        else:
            result += "missing (''). "
            result += "Note cannot be checked."
            status = "unchecked"

        definition = f"Check that the note is formatted properly for " \
                     f"{self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)
